import copy
import re
import traceback
from pathlib import Path

from loguru import logger

from sweepai.agents.assistant_wrapper import (
    client,
    openai_assistant_call,
    run_until_complete,
)
from sweepai.core.entities import AssistantRaisedException, FileChangeRequest, Message
from sweepai.logn.cache import file_cache
from sweepai.utils.chat_logger import ChatLogger, discord_log_error
from sweepai.utils.progress import AssistantConversation, TicketProgress

system_message = r""" You are searching through a codebase to guide a junior developer on how to solve the user request. The junior developer will follow your instructions exactly and make the changes.

# User Request
{user_request}

# Guide
## Step 1: Unzip the file into /mnt/data/repo. Then list all root level directories. You must copy the below code verbatim into the file.
```python
import zipfile
import os

zip_path = '{file_path}'
extract_to_path = 'mnt/data/repo'
os.makedirs(extract_to_path, exist_ok=True)
with zipfile.ZipFile(zip_path, 'r') as zip_ref:
    zip_ref.extractall(extract_to_path)
    zip_contents = zip_ref.namelist()
    root_dirs = {{name.split('/')[0] for name in zip_contents}}
    print(f'Root directories: {{root_dirs}}')
```

## Step 2: Find the relevant files.
You can search by file name or by keyword search in the contents.

## Step 3: Find relevant lines.
1. Locate the lines of code that contain the identified keywords or are at the specified line number. You can use keyword search or manually look through the file 100 lines at a time.
2. Check the surrounding lines to establish the full context of the code block.
3. Adjust the starting line to include the entire functionality that needs to be refactored or moved.
4. Finally determine the exact line spans that include a logical and complete section of code to be edited.

```python
def print_lines_with_keyword(content, keywords):
    max_matches=5
    context = 10

    matches = [i for i, line in enumerate(content.splitlines()) if any(keyword in line.lower() for keyword in keywords)]
    print(f"Found {{len(matches)}} matches, but capping at {{max_match}}")
    matches = matches[:max_matches]
    expanded_matches = set()

    for match in matches:
        start = max(0, match - context)
        end = min(len(content.splitlines()), match + context + 1)
        for i in range(start, end):
            expanded_matches.add(i)

    for i in sorted(expanded_matches):
        print(f"{{i}}: {{content.splitlines()[i]}}")
```

## Step 4: Construct a plan.
Provide the final plan to solve the issue, following these rules:
* DO NOT apply any changes here, they will not be persisted. You must provide the plan and the developer will apply the changes.
* You may only create new files and modify existing files.
* File paths should be relative paths from the root of the repo.
* Use the minimum number of create and modify operations required to solve the issue.
* Start and end lines indicate the exact start and end lines to edit. Expand this to encompass more lines if you're unsure where to make the exact edit.

Respond in the following format:

```xml
<plan>
<create_file file="file_path_1">
* Natural language instructions for creating the new file needed to solve the issue.
* Reference necessary files, imports and entity names.
...
</create_file>
...

<modify_file file="file_path_2" start_line="i" end_line="j">
* Natural language instructions for the modifications needed to solve the issue.
* Be concise and reference necessary files, imports and entity names.
...
</modify_file>
...

</plan>
```"""


@file_cache(ignore_params=["zip_path", "chat_logger", "ticket_progress"])
def new_planning(
    request: str,
    zip_path: str,
    additional_messages: list[Message] = [],
    chat_logger: ChatLogger | None = None,
    assistant_id: str = None,
    ticket_progress: TicketProgress | None = None,
) -> list[FileChangeRequest]:
    planning_iterations = 3
    try:

        def save_ticket_progress(assistant_id: str, thread_id: str, run_id: str):
            assistant_conversation = AssistantConversation.from_ids(
                assistant_id=assistant_id, run_id=run_id, thread_id=thread_id
            )
            if not assistant_conversation:
                return
            ticket_progress.planning_progress.assistant_conversation = (
                assistant_conversation
            )
            ticket_progress.save()

        logger.info("Uploading file...")
        zip_file_object = client.files.create(file=Path(zip_path), purpose="assistants")
        logger.info("Done uploading file.")
        zip_file_id = zip_file_object.id
        response = openai_assistant_call(
            request=request,
            assistant_id=assistant_id,
            additional_messages=additional_messages,
            uploaded_file_ids=[zip_file_id],
            chat_logger=chat_logger,
            save_ticket_progress=save_ticket_progress
            if ticket_progress is not None
            else None,
            instructions=system_message.format(
                user_request=request, file_path=f"mnt/data/{zip_file_id}"
            ),
        )
        run_id = response.run_id
        thread_id = response.thread_id
        for _ in range(planning_iterations):
            save_ticket_progress(
                assistant_id=response.assistant_id,
                thread_id=response.thread_id,
                run_id=response.run_id,
            )
            messages = response.messages
            final_message = messages.data[0].content[0].text.value
            fcrs = []
            fcr_matches = list(
                re.finditer(FileChangeRequest._regex, final_message, re.DOTALL)
            )
            if len(fcr_matches) > 0:
                break
            else:
                client.beta.threads.messages.create(
                    thread_id=thread_id,
                    role="user",
                    content="A valid plan (within the <plan> tags) was not provided. Please continue working on the plan. If you are stuck, consider starting over.",
                )
                run = client.beta.threads.runs.create(
                    thread_id=response.thread_id,
                    assistant_id=response.assistant_id,
                    instructions=system_message.format(
                        user_request=request, file_path=f"mnt/data/{zip_file_id}"
                    ),
                )
                run_id = run.id
                messages = run_until_complete(
                    thread_id=thread_id,
                    run_id=run_id,
                    assistant_id=response.assistant_id,
                )
        for match_ in fcr_matches:
            group_dict = match_.groupdict()
            if group_dict["change_type"] == "create_file":
                group_dict["change_type"] = "create"
            if group_dict["change_type"] == "modify_file":
                group_dict["change_type"] = "modify"
            fcr = FileChangeRequest(**group_dict)
            fcr.filename = fcr.filename.lstrip("/")
            fcr.instructions = fcr.instructions.replace("\n*", "\n•")
            fcr.instructions = fcr.instructions.strip("\n")
            if fcr.instructions.startswith("*"):
                fcr.instructions = "•" + fcr.instructions[1:]
            fcrs.append(fcr)
            new_file_change_request = copy.deepcopy(fcr)
            new_file_change_request.change_type = "check"
            new_file_change_request.parent = fcr
            fcrs.append(new_file_change_request)
        assert len(fcrs) > 0
        return fcrs
    except AssistantRaisedException as e:
        raise e
    except Exception as e:
        logger.exception(e)
        if chat_logger is not None:
            discord_log_error(
                str(e)
                + "\n\n"
                + traceback.format_exc()
                + "\n\n"
                + str(chat_logger.data)
            )
        return None


if __name__ == "__main__":
    request = """## Title: replace the broken tutorial link in installation.md with https://docs.sweep.dev/usage/tutorial\n"""
    additional_messages = [
        Message(
            role="user",
            content='<relevant_snippets_in_repo>\n<snippet source="docs/pages/usage/tutorial.mdx:45-60">\n...\n45: Now to be a Sweep power user, check out [Advanced: becoming a Sweep power user](https://docs.sweep.dev/usage/advanced).\n</snippet>\n<snippet source="docs/pages/usage/tutorial.mdx:30-45">\n...\n30: \n31: ![PR Comment](/tutorial/comment.png)\n32: \n33:     c. If you have GitHub Actions set up, it will automatically run the linters, build, and tests and will show any failed logs to Sweep to handle. This only works with GitHub Actions and not other CI providers, so unfortunately for Vercel we have to copy paste manually.\n34: \n35: ![GitHub Actions](/tutorial/github_actions.png)\n36: \n37: 6. Once you are happy with the PR, you can merge it and it will be deployed to production via Vercel.\n38: \n39: \n40: ![Final](/tutorial/final.png)\n41: \n42: \n43: You can see the final example at https://github.com/kevinlu1248/docusaurus-2/pull/4 with preview https://docusaurus-2-ql4cskc5o-sweepai.vercel.app/.\n44: \n45: Now to be a Sweep power user, check out [Advanced: becoming a Sweep power user](https://docs.sweep.dev/usage/advanced).\n...\n</snippet>\n<snippet source="docs/installation.md:45-60">\n...\n45: * Provide any additional context that might be helpful, e.g. see "src/App.test.tsx" for an example of a good unit test.\n46: * For more guidance, visit [Advanced](https://docs.sweep.dev/usage/advanced), or watch the following video.\n47: \n48: [![Video](http://img.youtube.com/vi/Qn9vB71R4UM/0.jpg)](http://www.youtube.com/watch?v=Qn9vB71R4UM "Advanced Sweep Tricks and Feedback Tips")\n49: \n50: For configuring Sweep for your repo, see [Config](https://docs.sweep.dev/usage/config), especially for setting up Sweep Rules and Sweep Sweep.\n51: \n52: ## Limitations of Sweep (for now) ⚠️\n53: \n54: * 🗃️ **Gigantic repos**: >5000 files. We have default extensions and directories to exclude but sometimes this doesn\'t catch them all. You may need to block some directories (see [`blocked_dirs`](https://docs.sweep.dev/usage/config#blocked_dirs))\n55:     * If Sweep is stuck at 0% for over 30 min and your repo has a few thousand files, let us know.\n56: \n57: * 🏗️ **Large-scale refactors**: >5 files or >300 lines of code changes (we\'re working on this!)\n58:     * We can\'t do this - "Refactor entire codebase from Tensorflow to PyTorch"\n59: \n60: * 🖼️ **Editing images** and other non-text assets\n...\n</snippet>\n<snippet source="docs/pages/usage/tutorial.mdx:0-15">\n0: # Tutorial for Getting Started with Sweep\n1: \n2: We recommend using an existing **real project** for Sweep, but if you must start from scratch, we recommend **using a template**. In particular, we recommend Vercel templates and Vercel auto-deploy, since Vercel\'s auto-generated previews make it **easy to review Sweep\'s PRs**\n3: \n4: We\'ll use [Docusaurus](https://vercel.com/templates/next.js/docusaurus-2) since it\'s is the easiest to set up (no backend). To see other templates see https://vercel.com/templates.\n5: \n6: 1. Go to https://vercel.com/templates/next.js/docusaurus-2 (or another template) and click "Deploy".\n7: \n8: ![Deploy](/tutorial/deployment.png)\n9: \n10: 2. Vercel will prompt you to select a GitHub account and click "Clone" after. This will trigger a build and deploy which will take a few minutes. Once the build is done, you will be greeted with a congratulations message.\n11: \n12: ![Congratulations](/tutorial/congratulations.png)\n13: \n14: 3. Go to the [Sweep Installation](https://github.com/apps/sweep-ai) page and click the grey "Configure" button or the green "Install" button. Ensure that that the Vercel template (i.e. Docusaurus) is configured to use Sweep.\n...\n</snippet>\n</relevant_snippets_in_repo>\ndocs/\n  installation.md\n  docs/pages/\n    docs/pages/usage/\n      _meta.json\n      advanced.mdx\n      config.mdx\n      extra-self-host.mdx\n      sandbox.mdx\n      tutorial.mdx',
            name=None,
            function_call=None,
            key=None,
        )
    ]
    print(
        new_planning(
            request,
            "/tmp/sweep_archive.zip",
            chat_logger=ChatLogger(
                {"username": "kevinlu1248", "title": "Unit test for planning"}
            ),
            ticket_progress=TicketProgress(tracking_id="ed47605a38"),
        )
    )



import traceback
from collections import OrderedDict

from sweepai.agents.assistant_function_modify import ensure_additional_messages_length, function_modify
from sweepai.core.entities import FileChangeRequest, MaxTokensExceeded, Message
from sweepai.logn.cache import file_cache
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.diff import generate_diff
from sweepai.utils.event_logger import logger
from sweepai.utils.github_utils import ClonedRepo, MockClonedRepo
from sweepai.utils.progress import AssistantConversation, TicketProgress


def is_blocked(file_path: str, blocked_dirs: list[str]):
    for blocked_dir in blocked_dirs:
        if file_path.startswith(blocked_dir) and len(blocked_dir) > 0:
            return {"success": True, "path": blocked_dir}
    return {"success": False}


def create_additional_messages(
    metadata: str,
    file_change_request: FileChangeRequest,
    changed_files: list[tuple[str, str]],
    comment_pr_diff_str: str,
    cloned_repo: ClonedRepo,
) -> tuple[list[Message], list[str]]:
    additional_messages = [
        Message(
            role="user",
            content=metadata,
            key="issue_metadata",
        )
    ]
    relevant_filepaths = []
    if comment_pr_diff_str and comment_pr_diff_str.strip():
        additional_messages = [
            Message(
                role="user",
                content="These changes have already been made:\n" + comment_pr_diff_str,
                key="pr_diffs",
            )
        ]
    file_path_to_contents = OrderedDict()
    # use only the latest change for each file
    # go forward to find the earliest version of each file in the array
    earliest_version_per_file = {}
    for file_path, (old_contents, new_contents) in changed_files:
        if file_path not in earliest_version_per_file:
            earliest_version_per_file[file_path] = old_contents
    latest_version_per_file = {}
    for file_path, (old_contents, new_contents) in reversed(changed_files):
        if file_path not in latest_version_per_file:
            latest_version_per_file[file_path] = new_contents
    for file_path, _ in changed_files:
        if (
            not latest_version_per_file[file_path]
            or not latest_version_per_file[file_path].strip()
        ):
            continue
        earliest_file_version = earliest_version_per_file[file_path]
        latest_file_version = latest_version_per_file[file_path]
        diffs = generate_diff(earliest_file_version, latest_file_version)
        if file_path not in file_path_to_contents:
            file_path_to_contents[file_path] = diffs
    changed_files_summary = "You have previously changed these files:\n" + "\n".join(
        [
            f'<changed_file file_path="{file_path}">\n{diffs}\n</changed_file>'
            for file_path, diffs in file_path_to_contents.items()
        ]
    )
    if changed_files:
        additional_messages += [
            Message(
                content=changed_files_summary,
                role="user",
                key="changed_files_summary",
            )
        ]
    if file_change_request.relevant_files:
        relevant_files_contents = []
        for file_path in file_change_request.relevant_files:
            try:
                relevant_files_contents.append(cloned_repo.get_file_contents(file_path))
            except Exception:
                for file_path, (old_contents, new_contents) in changed_files:
                    if file_path == file_path:
                        relevant_files_contents.append(new_contents)
                        break
                else:
                    relevant_files_contents.append("File not found")
        if relevant_files_contents:
            relevant_files_summary = "Relevant files in this PR:\n\n" + "\n".join(
                [
                    f'<relevant_file file_path="{file_path}">\n{file_contents}\n</relevant_file>'
                    for file_path, file_contents in zip(
                        file_change_request.relevant_files,
                        relevant_files_contents,
                    )
                ]
            )
            additional_messages.append(
                Message(
                    content=relevant_files_summary,
                    role="user",
                    key="relevant_files_summary",
                )
            )
            # keep all relevant_filepaths
            for file_path in file_change_request.relevant_files:
                relevant_filepaths.append(file_path)
    current_file_diff = ""
    if changed_files:
        for file_path, (old_contents, new_contents) in changed_files:
            if file_path == file_change_request.filename:
                current_file_diff += generate_diff(old_contents, new_contents) + "\n"
    return additional_messages, relevant_filepaths


@file_cache()
def modify_file(
    cloned_repo: ClonedRepo,
    metadata: str,
    file_change_request: FileChangeRequest,
    contents: str = "",
    branch: str = None,
    # context related
    changed_files: list[tuple[str, tuple[str, str]]] = [],
    comment_pr_diff_str: str = "",
    # o11y related
    assistant_conversation: AssistantConversation | None = None,
    ticket_progress: TicketProgress | None = None,
    chat_logger: ChatLogger | None = None,
    additional_messages: list[Message] = [],
):
    new_file = None
    try:
        relevant_file_messages, relevant_filepaths = create_additional_messages(
            metadata,
            file_change_request,
            changed_files,
            comment_pr_diff_str,
            cloned_repo,
        )
        additional_messages += relevant_file_messages
        additional_messages = ensure_additional_messages_length(additional_messages)
        new_file = function_modify(
            file_change_request.instructions,
            file_change_request.filename,
            contents or cloned_repo.get_file_contents(file_change_request.filename),
            additional_messages,
            chat_logger,
            start_line=file_change_request.start_line,
            end_line=file_change_request.end_line,
            ticket_progress=ticket_progress,
            assistant_conversation=assistant_conversation,
            relevant_filepaths=relevant_filepaths,
        )

    except Exception as e:  # Check for max tokens error
        if "max tokens" in str(e).lower():
            logger.error(f"Max tokens exceeded for {file_change_request.filename}")
            raise MaxTokensExceeded(file_change_request.filename)
        else:
            logger.error(f"Error: {e}")
            logger.error(traceback.format_exc())
            raise e
    return new_file


if __name__ == "__main__":
    cloned_repo = MockClonedRepo("/tmp/sweep", "sweepai/sweep")

    new_file = modify_file(
        cloned_repo=cloned_repo,
        metadata="This repo is Sweep.",
        file_change_request=FileChangeRequest(
            filename="push_image.sh",
            instructions="Add a print hello world statement before running anything.",
            change_type="modify",
        ),
    )
    print(new_file)



import json
import os
import re
import time
import traceback
from time import sleep
from typing import Callable, Optional

import openai
from anyio import Path
from loguru import logger
from openai.pagination import SyncCursorPage
from openai.types.beta.threads.thread_message import ThreadMessage
from openai.types.chat.chat_completion_message_tool_call import (
    ChatCompletionMessageToolCall,
    Function,
)
from pydantic import BaseModel

from sweepai.agents.assistant_functions import raise_error_schema
from sweepai.config.server import DEFAULT_GPT4_32K_MODEL, IS_SELF_HOSTED
from sweepai.core.entities import AssistantRaisedException, Message
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.event_logger import posthog
from sweepai.utils.openai_proxy import OpenAIProxy, get_client


def openai_retry_with_timeout(call, *args, num_retries=3, timeout=5, **kwargs):
    """
    Pass any OpenAI client call and retry it num_retries times, incorporating timeout into the call.

    Usage:
    run = openai_retry_with_timeout(client.beta.threads.runs.submit_tool_outputs, thread_id=thread.id, run_id=run.id, tool_outputs=tool_outputs, num_retries=3, timeout=10)

    Parameters:
    call (callable): The OpenAI client call to be retried.
    *args: Positional arguments for the callable.
    num_retries (int): The number of times to retry the call.
    timeout (int): The timeout value to be applied to the call.
    **kwargs: Keyword arguments for the callable.

    Returns:
    The result of the OpenAI client call.
    """
    error_message = None
    e = None
    for attempt in range(num_retries):
        try:
            return call(*args, **kwargs, timeout=timeout)
        except Exception as e_:
            logger.exception(f"Retry {attempt + 1} failed with error: {e}")
            error_message = str(e)
            e = e_
    if e is not None:
        raise Exception(
            f"Maximum retries reached. The call failed for call {error_message}"
        ) from error_message
    else:
        raise Exception(
            f"Maximum retries reached. The call failed for call {error_message}"
        )


def fix_tool_calls(tool_calls: Optional[list[ChatCompletionMessageToolCall]]):
    if tool_calls is None:
        return

    fixed_tool_calls = []

    for tool_call in tool_calls:
        current_function = tool_call.function.name
        try:
            function_args = json.loads(tool_call.function.arguments)
        except json.JSONDecodeError:
            logger.error(
                f"Error: could not decode function arguments: {tool_call.function.args}"
            )
            fixed_tool_calls.append(tool_call)
            continue
        if current_function in ("parallel", "multi_tool_use.parallel"):
            for _fake_i, _fake_tool_use in enumerate(function_args["tool_uses"]):
                _function_args = _fake_tool_use["parameters"]
                _current_function = _fake_tool_use["recipient_name"]
                if _current_function.startswith("functions."):
                    _current_function = _current_function[len("functions.") :]

                fixed_tc = ChatCompletionMessageToolCall(
                    id=f"{tool_call.id}_{_fake_i}",
                    type="function",
                    function=Function(
                        name=_current_function, arguments=json.dumps(_function_args)
                    ),
                )
                fixed_tool_calls.append(fixed_tc)
        else:
            fixed_tool_calls.append(tool_call)

    return fixed_tool_calls


save_ticket_progress_type = Callable[[str, str, str], None]


class AssistantResponse(BaseModel):
    messages: SyncCursorPage[ThreadMessage]
    assistant_id: str
    run_id: str
    thread_id: str


allowed_exts = [
    "c",
    "cpp",
    "csv",
    "docx",
    "html",
    "java",
    "json",
    "md",
    "pdf",
    "php",
    "pptx",
    "py",
    "rb",
    "tex",
    "txt",
    "css",
    "jpeg",
    "jpg",
    "js",
    "gif",
    "png",
    "tar",
    "ts",
    "xlsx",
    "xml",
    "zip",
    "elm",
]


def get_json_messages(
    thread_id: str,
    run_id: str,
    assistant_id: str,
):
    client = get_client()
    assistant = openai_retry_with_timeout(
        client.beta.assistants.retrieve,
        assistant_id=assistant_id,
    )
    messages = openai_retry_with_timeout(
        client.beta.threads.messages.list,
        thread_id=thread_id,
    )
    run_steps = openai_retry_with_timeout(
        client.beta.threads.runs.steps.list, run_id=run_id, thread_id=thread_id
    )
    system_message_json = {
        "role": "system",
        "content": assistant.instructions,
    }
    messages_json = [system_message_json]
    for message in messages:
        if message.role == "user":
            messages_json.append(
                {
                    "role": "user",
                    "content": message.content[0].text.value,
                }
            )
    for message_obj in list(run_steps.data)[:0:-1]:
        if message_obj.type == "message_creation":
            message_id = message_obj.step_details.message_creation.message_id
            thread_messages = openai_retry_with_timeout(
                client.beta.threads.messages.retrieve,
                message_id=message_id,
                thread_id=thread_id,
            )
            message_content = thread_messages.content[0].text.value
            messages_json.append(
                {
                    "role": "assistant",
                    "content": message_content,
                }
            )
            # TODO: handle annotations
        elif message_obj.type == "tool_calls":
            for tool_call in message_obj.step_details.tool_calls:
                if tool_call.type == "code_interpreter":
                    code_interpreter = tool_call.code_interpreter
                    input_ = code_interpreter.input
                    if not input_:
                        continue
                    input_content = f"Code interpreter input:\n```\n{input_}\n```"
                    messages_json.append(
                        {
                            "role": "assistant",
                            "content": input_content,
                        }
                    )
                    outputs = code_interpreter.outputs
                    output = outputs[0].logs if outputs else "__No output__"
                    output_content = f"Code interpreter output:\n```\n{output}\n```"
                    messages_json.append(
                        {
                            "role": "user",
                            "content": output_content,
                        }
                    )
                else:
                    function = tool_call.function
                    input_content = f"Function call of {function.name}:\n```\n{function.arguments}\n```"
                    messages_json.append(
                        {
                            "role": "assistant",
                            "content": input_content,
                        }
                    )
                    if function.output:
                        output_content = (
                            f"Function output:\n```\n{function.output}\n```"
                        )
                        messages_json.append(
                            {
                                "role": "user",
                                "content": output_content,
                            }
                        )
    return messages_json


def run_until_complete(
    thread_id: str,
    run_id: str,
    assistant_id: str,
    model: str = DEFAULT_GPT4_32K_MODEL,
    chat_logger: ChatLogger | None = None,
    sleep_time: int = 3,
    max_iterations: int = 2000,
    save_ticket_progress: save_ticket_progress_type | None = None,
):
    client = get_client()
    message_strings = []
    json_messages = []
    try:
        num_tool_calls_made = 0
        for i in range(max_iterations):
            run = openai_retry_with_timeout(
                client.beta.threads.runs.retrieve,
                thread_id=thread_id,
                run_id=run_id,
            )
            if run.status == "completed":
                logger.info(
                    f"Run completed with {run.status} (i={num_tool_calls_made})"
                )
                done_response = yield "done", {
                    "status": "completed",
                    "message": "Run completed successfully",
                }
                if not done_response:
                    break
                else:
                    run = client.beta.threads.runs.create(
                        thread_id=thread_id,
                        assistant_id=assistant_id,
                        instructions=done_response,
                        model=model,
                    )
            elif run.status in ("cancelled", "cancelling", "failed", "expired"):
                logger.info(
                    f"Run completed with {run.status} (i={num_tool_calls_made}) and reason {run.last_error}."
                )
                done_response = yield "done", {
                    "status": run.status,
                    "message": "Run failed",
                }
                if not done_response:
                    raise Exception(
                        f"Run failed assistant_id={assistant_id}, run_id={run_id}, thread_id={thread_id} with status {run.status} (i={num_tool_calls_made})"
                    )
                else:
                    run = client.beta.threads.runs.create(
                        thread_id=thread_id,
                        assistant_id=assistant_id,
                        instructions=done_response,
                        model=model,
                    )
            elif run.status == "requires_action":
                num_tool_calls_made += 1
                if num_tool_calls_made > 15 and model.startswith("gpt-3.5"):
                    raise AssistantRaisedException(
                        "Too many tool calls made on GPT 3.5."
                    )
                raw_tool_calls = [
                    tool_call
                    for tool_call in run.required_action.submit_tool_outputs.tool_calls
                ]
                tool_outputs = []
                tool_calls = []
                if any(
                    [
                        tool_call.function.name == raise_error_schema["name"]
                        for tool_call in raw_tool_calls
                    ]
                ):
                    arguments_parsed = json.loads(tool_calls[0].function.arguments)
                    raise AssistantRaisedException(arguments_parsed["message"])
                # tool_calls = raw_tool_calls
                for tool_call in raw_tool_calls:
                    try:
                        tool_call_arguments = re.sub(
                            r"\\+'", "", tool_call.function.arguments
                        )
                        function_input: dict = json.loads(tool_call_arguments)
                    except Exception:
                        logger.warning(
                            f"Could not parse function arguments (i={num_tool_calls_made}): {tool_call_arguments}"
                        )
                        tool_outputs.append(
                            {
                                "tool_call_id": tool_call.id,
                                "output": "FAILURE: Could not parse function arguments.",
                            }
                        )
                        continue
                    tool_function_name = tool_call.function.name
                    tool_function_input = function_input
                    # OpenAI has a bug where it calls the imaginary function "multi_tool_use.parallel"
                    # Based on https://github.com/phdowling/openai_multi_tool_use_parallel_patch/blob/main/openai_multi_tool_use_parallel_patch.py
                    if tool_function_name in ("multi_tool_use.parallel", "parallel"):
                        for fake_i, fake_tool_use in function_input["tool_uses"]:
                            function_input = fake_tool_use["parameters"]
                            function_name: str = fake_tool_use["recipient_name"]
                            function_name = function_name.removeprefix("functions.")
                            tool_calls.append(
                                (
                                    f"{tool_call.id}_{fake_i}",
                                    function_name,
                                    function_input,
                                )
                            )
                    else:
                        tool_calls.append(
                            (tool_call.id, tool_function_name, tool_function_input)
                        )

                for tool_call_id, tool_function_name, tool_function_input in tool_calls:
                    tool_output = yield tool_function_name, tool_function_input
                    tool_output_formatted = {
                        "tool_call_id": tool_call_id,
                        "output": tool_output,
                    }
                    tool_outputs.append(tool_output_formatted)
                run = openai_retry_with_timeout(
                    client.beta.threads.runs.submit_tool_outputs,
                    thread_id=thread_id,
                    run_id=run.id,
                    tool_outputs=tool_outputs,
                )
            if save_ticket_progress is not None:
                save_ticket_progress(
                    assistant_id=assistant_id,
                    thread_id=thread_id,
                    run_id=run_id,
                )
            messages = openai_retry_with_timeout(
                client.beta.threads.messages.list,
                thread_id=thread_id,
            )
            current_message_strings = [
                message.content[0].text.value if message.content else ""
                for message in messages.data
            ]
            if message_strings != current_message_strings and current_message_strings:
                logger.info(run.status)
                logger.info(current_message_strings[0])
                message_strings = current_message_strings
                json_messages = get_json_messages(
                    thread_id=thread_id,
                    run_id=run_id,
                    assistant_id=assistant_id,
                )
                if chat_logger is not None:
                    chat_logger.add_chat(
                        {
                            "model": model,
                            "messages": json_messages,
                            "output": message_strings[0],
                            "thread_id": thread_id,
                            "run_id": run_id,
                            "max_tokens": 1000,
                            "temperature": 0,
                        }
                    )
            else:
                if i % 5 == 0:
                    logger.info(run.status)
            if i == max_iterations - 1:
                logger.warning(
                    f"run_until_complete hit max iterations, run.status is {run.status}"
                )
            time.sleep(sleep_time)
    except (KeyboardInterrupt, SystemExit):
        client.beta.threads.runs.cancel(thread_id=thread_id, run_id=run_id)
        logger.warning(f"Run cancelled: {run_id} (n={num_tool_calls_made})")
        raise SystemExit
    if save_ticket_progress is not None:
        save_ticket_progress(
            assistant_id=assistant_id,
            thread_id=thread_id,
            run_id=run_id,
        )
    for json_message in json_messages:
        logger.info(f'(n={num_tool_calls_made}) {json_message["content"]}')
    return client.beta.threads.messages.list(
        thread_id=thread_id,
    )


def openai_assistant_call_helper(
    request: str,
    instructions: str | None = None,
    additional_messages: list[Message] = [],
    file_paths: list[str] = [],  # use either file_paths or file_ids
    uploaded_file_ids: list[str] = [],
    tools: list[dict[str, str]] = [{"type": "code_interpreter"}],
    model: str = DEFAULT_GPT4_32K_MODEL,
    sleep_time: int = 3,
    chat_logger: ChatLogger | None = None,
    assistant_id: str | None = None,
    assistant_name: str | None = None,
    save_ticket_progress: save_ticket_progress_type | None = None,
):
    file_ids = [] if not uploaded_file_ids else uploaded_file_ids
    file_object = None
    client = get_client()
    if not file_ids:
        for file_path in file_paths:
            if not any(file_path.endswith(extension) for extension in allowed_exts):
                os.rename(file_path, file_path + ".txt")
                file_path += ".txt"
            file_object = client.files.create(
                file=Path(file_path), purpose="assistants"
            )
            file_ids.append(file_object.id)

    logger.debug(instructions)
    # always create new one
    assistant = openai_retry_with_timeout(
        client.beta.assistants.create,
        name=assistant_name,
        instructions=instructions,
        tools=tools,
        model=model,
    )
    thread = client.beta.threads.create()
    if file_ids:
        logger.info("Uploading files...")
    if request:
        client.beta.threads.messages.create(
            thread_id=thread.id,
            role="user",
            content=request,
            file_ids=file_ids,
        )
    if file_ids:
        logger.info("Files uploaded")
    for message in additional_messages:
        client.beta.threads.messages.create(
            thread_id=thread.id,
            role="user",
            content=message.content,
        )
    run = client.beta.threads.runs.create(
        thread_id=thread.id,
        assistant_id=assistant.id,
        instructions=instructions,
        model=model,
    )
    if len(tools) > 1:
        return run_until_complete(
            thread_id=thread.id,
            run_id=run.id,
            model=model,
            chat_logger=chat_logger,
            assistant_id=assistant.id,
            sleep_time=sleep_time,
            save_ticket_progress=save_ticket_progress,
        )
    for file_id in file_ids:
        client.files.delete(file_id=file_id)
    return (
        assistant.id,
        run.id,
        thread.id,
    )


# Split in two so it can be cached
def openai_assistant_call(
    request: str,
    instructions: str | None = None,
    additional_messages: list[Message] = [],
    file_paths: list[str] = [],
    uploaded_file_ids: list[str] = [],
    tools: list[dict[str, str]] = [{"type": "code_interpreter"}],
    model: str = DEFAULT_GPT4_32K_MODEL,
    sleep_time: int = 3,
    chat_logger: ChatLogger | None = None,
    assistant_id: str | None = None,
    assistant_name: str | None = None,
    save_ticket_progress: save_ticket_progress_type | None = None,
):
    client = get_client()
    model = (
        "gpt-3.5-turbo-1106"
        if (chat_logger is None or chat_logger.use_faster_model())
        and not IS_SELF_HOSTED
        else DEFAULT_GPT4_32K_MODEL
    )
    posthog.capture(
        chat_logger.data.get("username") if chat_logger is not None else "anonymous",
        "call_assistant_api",
        {
            "query": request,
            "model": model,
            "username": (
                chat_logger.data.get("username", "anonymous")
                if chat_logger is not None
                else "anonymous"
            ),
            "is_self_hosted": IS_SELF_HOSTED,
            "trace": "".join(traceback.format_list(traceback.extract_stack())),
        },
    )
    retries = range(3)
    for _ in retries:
        try:
            response = openai_assistant_call_helper(
                request=request,
                instructions=instructions,
                additional_messages=additional_messages,
                file_paths=file_paths,
                uploaded_file_ids=uploaded_file_ids,
                tools=tools,
                model=model,
                sleep_time=sleep_time,
                chat_logger=chat_logger,
                assistant_id=assistant_id,
                assistant_name=assistant_name,
                save_ticket_progress=save_ticket_progress,
            )
            if len(tools) > 1:
                return response
            (assistant_id, run_id, thread_id) = response
            messages = client.beta.threads.messages.list(
                thread_id=thread_id,
            )
            return AssistantResponse(
                messages=messages,
                assistant_id=assistant_id,
                run_id=run_id,
                thread_id=thread_id,
            )
        except AssistantRaisedException as e:
            logger.warning(e.message)
        except Exception as e:
            logger.exception(e)
            raise e


def run_until_complete_unstable(
    tools: list[dict[str, str]],
    model: str = DEFAULT_GPT4_32K_MODEL,
    chat_logger: ChatLogger | None = None,
    sleep_time: int = 3,
    max_iterations: int = 100,
    save_ticket_progress: save_ticket_progress_type | None = None,
    messages: list[Message] = [],
):
    get_client()
    # used for chat logger
    for i in range(max_iterations):
        # log our progress
        if i % 5 == 0:
            logger.info(
                f"run_until_complete iteration {i}, current message length: {len(messages)}"
            )
        # if we are somehow about to hit the max iterations, log a warning
        if i == max_iterations - 1:
            logger.warning(
                f"run_until_complete about to hit max iterations! iteration {i} out {max_iterations}"
            )
        # get the response from openai
        try:
            openai_proxy = OpenAIProxy()
            response = openai_proxy.call_openai(
                model,
                messages,
                tools,
                max_tokens=2048,
                temperature=0.2,
                # set max tokens later
            )
        # sometimes deployment for opennai is not found, retry after a minute
        except openai.NotFoundError as e:
            logger.error(
                f"Openai deployment not found on iteration {i} with error: {e}\n Retrying in 60 seconds..."
            )
            sleep(60)
            continue
        except Exception as e:
            logger.error(f"chat completions failed on interation {i} with error: {e}")
            sleep(sleep_time)
            continue

        response_message = response.choices[0].message
        tool_calls = fix_tool_calls(response_message.tool_calls)
        response_message.tool_calls = tool_calls
        # extend conversation
        response_message_dict = response_message.dict()
        # in some cases the fields are None and we must replace these with empty strings
        for key, value in response_message_dict.items():
            if value is None:
                response_message_dict[key] = ""
        # if function_call is None we must remove it or else openai will throw an error
        if response_message_dict.get("function_call", "not in dict") == "":
            response_message_dict.pop("function_call")
        if response_message_dict.get("tool_calls", "not in dict") == "":
            response_message_dict.pop("tool_calls")

        messages.append(response_message_dict)
        # if a tool call was made
        if tool_calls:
            for tool_call in tool_calls:
                function_name = tool_call.function.name
                try:
                    function_args = json.loads(tool_call.function.arguments)
                except json.JSONDecodeError as e:
                    logger.debug(
                        f"Error: could not decode function arguments: {tool_call.function.args}"
                    )
                    tool_output = f"ERROR\nCould not decode function arguments:\n{e}"
                else:
                    logger.debug(
                        f"tool_call: {function_name} with args: {function_args}"
                    )
                    tool_output = yield function_name, function_args
                messages.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": function_name,
                        "content": tool_output,
                    }
                )  # extend conversation with function response
                if not tool_output:
                    break
        else:  # no tool call being made implies either an error or a success
            logger.info(
                f"no tool calls were made, we are done - message: {response_message}"
            )
            done_response = yield "done", {
                "status": "completed",
                "message": "Run completed successfully",
            }
            logger.info(
                f"run_until_complete done_response: {done_response} completed after {i} iterations"
            )
            if not done_response:
                break

        # on each iteration of the for loop, we will log to chat_logger
        if chat_logger is not None and len(messages):
            descriptive_messages = [message for message in messages]
            # for tool calls, the content is empty, replace that with the function contents
            for message in descriptive_messages:
                if message.get("content", "") == "" and "tool_calls" in message:
                    # if there were multiple tool calls, add them both
                    for tool_call in message["tool_calls"]:
                        message[
                            "content"
                        ] += f"\n\ntool_call: {tool_call.get('function', '')}"
            chat_logger.add_chat(
                {
                    "model": model,
                    "messages": descriptive_messages,
                    "output": descriptive_messages[-1]["content"],
                    "max_tokens": 1000,
                    "temperature": 0,
                }
            )
        # update ticket_progress
        # if save_ticket_progress is not None:
        #     save_ticket_progress(
        #         messages=messages
        #     )


def openai_assistant_call_helper_unstable(
    request: str,
    instructions: str | None = None,
    additional_messages: list[Message] = [],
    file_paths: list[str] = [],  # use either file_paths or file_ids
    uploaded_file_ids: list[str] = [],
    tools: list[dict[str, str]] = [{"type": "code_interpreter"}],
    model: str = DEFAULT_GPT4_32K_MODEL,
    sleep_time: int = 3,
    chat_logger: ChatLogger | None = None,
    assistant_id: str | None = None,
    assistant_name: str | None = None,
    save_ticket_progress: save_ticket_progress_type | None = None,
):
    get_client()
    logger.debug(instructions)
    messages = [{"role": "system", "content": instructions}]
    for message in additional_messages:
        messages.append({"role": message.role, "content": message.content})

    # tools must always be > 1
    if len(tools) > 1:
        return run_until_complete_unstable(
            tools=tools,
            messages=messages,
            model=model,
            chat_logger=chat_logger,
            sleep_time=sleep_time,
            save_ticket_progress=save_ticket_progress,
        )
    else:
        raise Exception("openai_assistant_call_helper tools must be > 1")


# Split in two so it can be cached
def openai_assistant_call_unstable(
    request: str,
    instructions: str | None = None,
    additional_messages: list[Message] = [],
    file_paths: list[str] = [],
    uploaded_file_ids: list[str] = [],
    tools: list[dict[str, str]] = [{"type": "code_interpreter"}],
    model: str = DEFAULT_GPT4_32K_MODEL,
    sleep_time: int = 3,
    chat_logger: ChatLogger | None = None,
    assistant_id: str | None = None,
    assistant_name: str | None = None,
    save_ticket_progress: save_ticket_progress_type | None = None,
):
    client = get_client()
    if chat_logger and chat_logger.use_faster_model():
        raise Exception("GPT-3.5 is not supported on assistant calls.")
    model = DEFAULT_GPT4_32K_MODEL
    posthog.capture(
        chat_logger.data.get("username") if chat_logger is not None else "anonymous",
        "call_assistant_api",
        {
            "query": request,
            "model": model,
            "username": (
                chat_logger.data.get("username", "anonymous")
                if chat_logger is not None
                else "anonymous"
            ),
            "is_self_hosted": IS_SELF_HOSTED,
            "trace": "".join(traceback.format_list(traceback.extract_stack())),
        },
    )
    retries = range(3)
    for _ in retries:
        try:
            response = openai_assistant_call_helper_unstable(
                request=request,
                instructions=instructions,
                additional_messages=additional_messages,
                file_paths=file_paths,
                uploaded_file_ids=uploaded_file_ids,
                tools=tools,
                model=model,
                sleep_time=sleep_time,
                chat_logger=chat_logger,
                assistant_id=assistant_id,
                assistant_name=assistant_name,
                save_ticket_progress=save_ticket_progress,
            )
            if len(tools) > 1:
                return response
            (assistant_id, run_id, thread_id) = response
            messages = client.beta.threads.messages.list(
                thread_id=thread_id,
            )
            return AssistantResponse(
                messages=messages,
                assistant_id=assistant_id,
                run_id=run_id,
                thread_id=thread_id,
            )
        except AssistantRaisedException as e:
            logger.warning(e.message)
        except Exception as e:
            logger.error(e)
            raise e



from __future__ import annotations

import ast
import os
import re
import traceback
import uuid
from dataclasses import dataclass
from io import StringIO
from typing import Optional

import tiktoken
from loguru import logger
from pylint.lint import Run
from pylint.reporters.text import TextReporter
from tree_sitter import Node
from tree_sitter_languages import get_parser

from sweepai.core.entities import Snippet
from sweepai.logn.cache import file_cache
from sweepai.utils.chat_logger import discord_log_error


def non_whitespace_len(s: str) -> int:  # new len function
    return len(re.sub("\s", "", s))


def get_line_number(index: int, source_code: str) -> int:
    total_chars = 0
    line_number = 0
    for line_number, line in enumerate(source_code.splitlines(keepends=True), start=1):
        total_chars += len(line)
        if total_chars > index:
            return line_number - 1
    return line_number


@dataclass
class Span:
    # Represents a slice of a string
    start: int = 0
    end: int = 0

    def __post_init__(self):
        # If end is None, set it to start
        if self.end is None:
            self.end = self.start

    def extract(self, s: str) -> str:
        # Grab the corresponding substring of string s by bytes
        return s[self.start : self.end]

    def extract_lines(self, s: str) -> str:
        # Grab the corresponding substring of string s by lines
        return "\n".join(s.splitlines()[self.start : self.end + 1])

    def __add__(self, other: Span | int) -> Span:
        # e.g. Span(1, 2) + Span(2, 4) = Span(1, 4) (concatenation)
        # There are no safety checks: Span(a, b) + Span(c, d) = Span(a, d)
        # and there are no requirements for b = c.
        if isinstance(other, int):
            return Span(self.start + other, self.end + other)
        elif isinstance(other, Span):
            return Span(self.start, other.end)
        else:
            raise NotImplementedError()

    def __len__(self) -> int:
        # i.e. Span(a, b) = b - a
        return self.end - self.start


AVG_CHAR_IN_LINE = 60


def chunk_tree(
    tree,
    source_code: bytes,
    MAX_CHARS=AVG_CHAR_IN_LINE * 200,  # 200 lines of code
    coalesce=AVG_CHAR_IN_LINE * 50,  # 50 lines of code
) -> list[Span]:
    # 1. Recursively form chunks based on the last post (https://docs.sweep.dev/blogs/chunking-2m-files)
    def chunk_node(node: Node) -> list[Span]:
        chunks: list[Span] = []
        current_chunk: Span = Span(node.start_byte, node.start_byte)
        node_children = node.children
        for child in node_children:
            if child.end_byte - child.start_byte > MAX_CHARS:
                chunks.append(current_chunk)
                current_chunk = Span(child.end_byte, child.end_byte)
                chunks.extend(chunk_node(child))
            elif child.end_byte - child.start_byte + len(current_chunk) > MAX_CHARS:
                chunks.append(current_chunk)
                current_chunk = Span(child.start_byte, child.end_byte)
            else:
                current_chunk += Span(child.start_byte, child.end_byte)
        chunks.append(current_chunk)
        return chunks

    chunks = chunk_node(tree.root_node)

    # 2. Filling in the gaps
    if len(chunks) == 0:
        return []
    if len(chunks) < 2:
        end = get_line_number(chunks[0].end, source_code)
        return [Span(0, end)]
    for i in range(len(chunks) - 1):
        chunks[i].end = chunks[i + 1].start
    chunks[-1].end = tree.root_node.end_byte

    # 3. Combining small chunks with bigger ones
    new_chunks = []
    current_chunk = Span(0, 0)
    for chunk in chunks:
        current_chunk += chunk
        if non_whitespace_len(
            current_chunk.extract(source_code.decode("utf-8"))
        ) > coalesce and "\n" in current_chunk.extract(source_code.decode("utf-8")):
            new_chunks.append(current_chunk)
            current_chunk = Span(chunk.end, chunk.end)
    if len(current_chunk) > 0:
        new_chunks.append(current_chunk)

    # 4. Changing line numbers
    first_chunk = new_chunks[0]
    line_chunks = [Span(0, get_line_number(first_chunk.end, source_code))]
    for chunk in new_chunks[1:]:
        start_line = get_line_number(chunk.start, source_code) + 1
        end_line = get_line_number(chunk.end, source_code)
        line_chunks.append(Span(start_line, max(start_line, end_line)))

    # 5. Eliminating empty chunks
    line_chunks = [chunk for chunk in line_chunks if len(chunk) > 0]

    # 6. Coalescing last chunk if it's too small
    if len(line_chunks) > 1 and len(line_chunks[-1]) < coalesce:
        line_chunks[-2] += line_chunks[-1]
        line_chunks.pop()

    return line_chunks


extension_to_language = {
    "js": "tsx",
    "jsx": "tsx",
    "ts": "tsx",
    "tsx": "tsx",
    "mjs": "tsx",
    "py": "python",
    "rs": "rust",
    "go": "go",
    "java": "java",
    "cpp": "cpp",
    "cc": "cpp",
    "cxx": "cpp",
    "c": "cpp",
    "h": "cpp",
    "hpp": "cpp",
    "cs": "cpp",
    "rb": "ruby",
    # "md": "markdown",
    # "rst": "markdown",
    # "txt": "markdown",
    # "erb": "embedded-template",
    # "ejs": "embedded-template",
    # "html": "embedded-template",
    "erb": "html",
    "ejs": "html",
    "html": "html",
    "vue": "html",
    "php": "php",
    "elm": "elm",
}


def naive_chunker(code: str, line_count: int = 30, overlap: int = 15):
    if overlap >= line_count:
        raise ValueError("Overlap should be smaller than line_count.")
    lines = code.split("\n")
    total_lines = len(lines)
    chunks = []

    start = 0
    while start < total_lines:
        end = min(start + line_count, total_lines)
        chunk = "\n".join(lines[start:end])
        chunks.append(chunk)
        start += line_count - overlap

    return chunks


def check_valid_typescript(code: str) -> tuple[bool, str]:
    # with tempfile.TemporaryDirectory() as temp_dir:
    #     file_hash = uuid.uuid4().hex[:10]
    #     tmp_file = os.path.join(temp_dir, file_hash + "_" + "temp.ts")

    #     with open(tmp_file, "w") as file:
    #         file.write(code)

    #     result = subprocess.run(
    #         ["npx", "prettier", "--parser", "babel-ts", tmp_file],
    #         capture_output=True,
    #         timeout=5,
    #     )

    #     os.remove(tmp_file)
    #     return result.returncode == 0, (result.stdout + result.stderr).decode("utf-8")
    return True, ""


def check_syntax(file_path: str, code: str) -> tuple[bool, str]:
    ext = file_path.split(".")[-1]
    if ext in extension_to_language:
        language = extension_to_language[ext]
    else:
        return True, "Unsupported file extension, skipping syntax check."
    parser = get_parser(language)
    tree = parser.parse(code.encode("utf-8"))

    if language == "python":
        # First check for syntax errors
        try:
            ast.parse(code)
        except SyntaxError as e:
            error_message = f"Python syntax error: {e.msg} at line {e.lineno}"
            return False, error_message

    if ext in ("tsx", "ts"):
        is_valid, error_message = check_valid_typescript(code)
        if not is_valid:
            return is_valid, error_message

    def find_deepest_error(node: Node) -> Optional[Node]:
        deepest_error = None
        if node.has_error:
            deepest_error = node
        for child in node.children:
            child_error = find_deepest_error(child)
            if child_error:
                deepest_error = child_error
        return deepest_error

    error_location = find_deepest_error(tree.root_node)
    if error_location:
        line_number, _ = error_location.start_point
        error_start = max(0, line_number - 20)
        error_span = "\n".join(code.split("\n")[error_start : line_number + 10])
        error_message = f"Invalid syntax found around lines {error_start}-{line_number}, displayed below:\n{error_span}"
        return (False, error_message)
    return True, ""


def check_code(file_path: str, code: str) -> tuple[bool, str]:
    is_valid, error_message = check_syntax(file_path, code)
    if not is_valid:
        return is_valid, error_message
    ext = file_path.split(".")[-1]
    if ext == "py":
        file_hash = uuid.uuid4().hex
        new_file = os.path.join("/tmp", file_hash + "_" + os.path.basename(file_path))
        try:
            with open(new_file, "w") as f:
                f.write(code)
            pylint_output = StringIO()
            reporter = TextReporter(pylint_output)
            Run(
                [
                    new_file,
                    "--errors-only",
                    "--disable=import-error",
                    "--disable=no-member",
                    "--disable=relative-beyond-top-level",
                ],
                reporter=reporter,
                do_exit=False,
            )
            error_message = pylint_output.getvalue()
            try:
                os.remove(new_file)
            except FileNotFoundError:
                pass
            if error_message:
                return False, error_message
        except Exception as e:
            discord_log_error("Pylint BS:\n" + e + traceback.format_exc())
    return True, ""


@file_cache()
def chunk_code(
    code: str,
    path: str,
    MAX_CHARS=AVG_CHAR_IN_LINE * 200,  # 200 lines of code
    coalesce=80,
) -> list[Snippet]:
    ext = path.split(".")[-1]
    if ext in extension_to_language:
        language = extension_to_language[ext]
    else:
        # Fallback to naive chunking if tree_sitter fails
        line_count = 50
        overlap = 0
        chunks = naive_chunker(code, line_count, overlap)
        snippets = []
        for idx, chunk in enumerate(chunks):
            end = min((idx + 1) * (line_count - overlap), len(code.split("\n")))
            new_snippet = Snippet(
                content=code,
                start=idx * (line_count - overlap),
                end=end,
                file_path=path,
            )
            snippets.append(new_snippet)
        return snippets
    try:
        parser = get_parser(language)
        tree = parser.parse(code.encode("utf-8"))
        chunks = chunk_tree(
            tree, code.encode("utf-8"), MAX_CHARS=MAX_CHARS, coalesce=coalesce
        )
        snippets = []
        for chunk in chunks:
            new_snippet = Snippet(
                content=code,
                start=chunk.start,
                end=chunk.end,
                file_path=path,
            )
            snippets.append(new_snippet)
        return snippets
    except SystemExit:
        raise SystemExit
    except Exception:
        logger.error(traceback.format_exc())
        return []


TIKTOKEN_CACHE_DIR = "/tmp/cache/tiktoken"


class Tiktoken:
    def __init__(self):
        openai_models = [
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-1106",
            "gpt-4",
            "gpt-4-32k",
            "gpt-4-32k-0613",
            "gpt-4-1106-preview",
            "gpt-4-0125-preview",
        ]
        self.openai_models = {
            model: tiktoken.encoding_for_model(model) for model in openai_models
        }

    def count(self, text: str, model: str = "gpt-4") -> int:
        return len(self.openai_models[model].encode(text, disallowed_special=()))

    def truncate_string(
        self, text: str, model: str = "gpt-4", max_tokens: int = 8192
    ) -> str:
        tokens = self.openai_models[model].encode(text)[:max_tokens]
        return self.openai_models[model].decode(tokens)


test_code = """
import React from 'react';
import { render } from '@testing-library/react';
import CallToAction from '../components/CallToAction';
describe('CallToAction component', () => {
  it('renders the correct YouTube video link', () => {
    const { getByTitle } = render(<CallToAction />);
    const iframeElement = getByTitle('YouTube video player');
    expect(iframeElement.getAttribute('src')).toBe('https://www.youtube.com/embed/GVEkDZmWw8E?autoplay=1&mute=1&loop=1&vq=hd1080&modestbranding=1&controls=0');
    it('has a button with the correct color properties', () => {
    const { getByRole } = render(<CallToAction />);
    const buttonElement = getByRole('button', { name: /install sweep/i });
    expect(buttonElement).toHaveStyle({
      colorScheme: 'green',
      bg: 'green.400',
      _hover: { bg: 'green.600' }
    });
  });
});
"""


if __name__ == "__main__":
    print(check_syntax("CallToAction.test.tsx", test_code))



import traceback
from time import time

from loguru import logger

from sweepai.config.client import SweepConfig, get_blocked_dirs
from sweepai.core.context_pruning import RepoContextManager, get_relevant_context
from sweepai.core.lexical_search import (
    compute_vector_search_scores,
    prepare_lexical_search_index,
    search_index,
)
from sweepai.logn.cache import file_cache
from sweepai.utils.chat_logger import discord_log_error
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import ClonedRepo
from sweepai.utils.progress import TicketProgress


@file_cache()
def get_top_k_snippets(
    cloned_repo: ClonedRepo,
    query: str,
    ticket_progress: TicketProgress | None = None,
    k: int = 7,
):
    sweep_config: SweepConfig = SweepConfig()
    blocked_dirs = get_blocked_dirs(cloned_repo.repo)
    sweep_config.exclude_dirs += blocked_dirs
    _, snippets, lexical_index = prepare_lexical_search_index(
        cloned_repo.cached_dir,
        sweep_config,
        ticket_progress,
        ref_name=f"{str(cloned_repo.git_repo.head.commit.hexsha)}",
    )
    if ticket_progress:
        ticket_progress.search_progress.indexing_progress = (
            ticket_progress.search_progress.indexing_total
        )
        ticket_progress.save()

    for snippet in snippets:
        snippet.file_path = snippet.file_path[len(cloned_repo.cached_dir) + 1 :]
    content_to_lexical_score = search_index(query, lexical_index)
    files_to_scores = compute_vector_search_scores(query, snippets)
    for snippet in snippets:
        vector_score = files_to_scores.get(snippet.denotation, 0.04)
        snippet_score = 0.02
        if snippet.denotation in content_to_lexical_score:
            # roughly fine tuned vector score weight based on average score from search_eval.py on 10 test cases Feb. 13, 2024
            snippet_score = content_to_lexical_score[snippet.denotation] + (
                vector_score * 3.5
            )
            content_to_lexical_score[snippet.denotation] = snippet_score
        else:
            content_to_lexical_score[snippet.denotation] = snippet_score * vector_score

    ranked_snippets = sorted(
        snippets,
        key=lambda snippet: content_to_lexical_score[snippet.denotation],
        reverse=True,
    )
    ranked_snippets = ranked_snippets[:k]
    return ranked_snippets, snippets, content_to_lexical_score


def prep_snippets(
    cloned_repo: ClonedRepo,
    query: str,
    ticket_progress: TicketProgress | None = None,
    k: int = 7,
):
    ranked_snippets, snippets, content_to_lexical_score = get_top_k_snippets(
        cloned_repo, query, ticket_progress, k
    )
    if ticket_progress:
        ticket_progress.search_progress.retrieved_snippets = ranked_snippets
        ticket_progress.save()
    snippet_paths = [snippet.file_path for snippet in ranked_snippets]
    prefixes = []
    for snippet_path in snippet_paths:
        snippet_depth = len(snippet_path.split("/"))
        for idx in range(snippet_depth):  # heuristic
            if idx > snippet_depth // 2:
                prefixes.append("/".join(snippet_path.split("/")[:idx]) + "/")
        prefixes.append(snippet_path)
    _, dir_obj = cloned_repo.list_directory_tree(
        included_directories=list(set(prefixes)),
        included_files=list(set(snippet_paths)),
    )
    repo_context_manager = RepoContextManager(
        dir_obj=dir_obj,
        current_top_tree=str(dir_obj),
        current_top_snippets=ranked_snippets,
        snippets=snippets,
        snippet_scores=content_to_lexical_score,
        cloned_repo=cloned_repo,
    )
    return repo_context_manager


def fetch_relevant_files(
    cloned_repo,
    title,
    summary,
    replies_text,
    username,
    metadata,
    on_ticket_start_time,
    tracking_id,
    is_paying_user,
    is_consumer_tier,
    issue_url,
    chat_logger,
    ticket_progress: TicketProgress,
):
    logger.info("Fetching relevant files...")
    try:
        search_query = (title + summary + replies_text).strip("\n")
        replies_text = f"\n{replies_text}" if replies_text else ""
        formatted_query = (f"{title.strip()}\n{summary.strip()}" + replies_text).strip(
            "\n"
        )
        repo_context_manager = prep_snippets(cloned_repo, search_query, ticket_progress)
        ticket_progress.search_progress.repo_tree = str(repo_context_manager.dir_obj)
        ticket_progress.save()

        repo_context_manager = get_relevant_context(
            formatted_query,
            repo_context_manager,
            ticket_progress,
            chat_logger=chat_logger,
        )
        snippets = repo_context_manager.current_top_snippets
        ticket_progress.search_progress.repo_tree = str(repo_context_manager.dir_obj)
        ticket_progress.search_progress.final_snippets = snippets
        ticket_progress.save()

        tree = str(repo_context_manager.dir_obj)
        dir_obj = repo_context_manager.dir_obj
    except Exception as e:
        trace = traceback.format_exc()
        logger.exception(f"{trace} (tracking ID: `{tracking_id}`)")
        log_error(
            is_paying_user,
            is_consumer_tier,
            username,
            issue_url,
            "File Fetch",
            str(e) + "\n" + traceback.format_exc(),
            priority=1,
        )
        posthog.capture(
            username,
            "failed",
            properties={
                **metadata,
                "error": str(e),
                "duration": time() - on_ticket_start_time,
            },
        )
        raise e
    return snippets, tree, dir_obj


SLOW_MODE = False
SLOW_MODE = True


def log_error(
    is_paying_user,
    is_trial_user,
    username,
    issue_url,
    error_type,
    exception,
    priority=0,
):
    if is_paying_user or is_trial_user:
        if priority == 1:
            priority = 0
        elif priority == 2:
            priority = 1

    prefix = ""
    if is_trial_user:
        prefix = " (TRIAL)"
    if is_paying_user:
        prefix = " (PRO)"

    content = (
        f"**{error_type} Error**{prefix}\n{username}:"
        f" {issue_url}\n```{exception}```"
    )
    discord_log_error(content, priority=2)


def center(text: str) -> str:
    return f"<div align='center'>{text}</div>"


def fire_and_forget_wrapper(call):
    """
    This decorator is used to run a function in a separate thread.
    It does not return anything and does not wait for the function to finish.
    It fails silently.
    """

    def wrapper(*args, **kwargs):
        try:
            return call(*args, **kwargs)
        except Exception:
            pass
        # def run_in_thread(call, *a, **kw):
        #     try:
        #         call(*a, **kw)
        #     except:
        #         pass

        # thread = Thread(target=run_in_thread, args=(call,) + args, kwargs=kwargs)
        # thread.start()

    return wrapper



"""
on_ticket is the main function that is called when a new issue is created.
It is only called by the webhook handler in sweepai/api.py.
"""

import difflib
import io
import os
import re
import traceback
import zipfile
from copy import deepcopy
from time import time

import markdown
import openai
import requests
import yaml
import yamllint.config as yamllint_config
from github import BadCredentialsException, Github, Repository
from github.Issue import Issue
from github.PullRequest import PullRequest as GithubPullRequest
from loguru import logger
from tabulate import tabulate
from tqdm import tqdm
from yamllint import linter

from sweepai.agents.pr_description_bot import PRDescriptionBot
from sweepai.config.client import (
    DEFAULT_RULES,
    RESET_FILE,
    RESTART_SWEEP_BUTTON,
    REVERT_CHANGED_FILES_TITLE,
    RULES_LABEL,
    RULES_TITLE,
    SWEEP_BAD_FEEDBACK,
    SWEEP_GOOD_FEEDBACK,
    SweepConfig,
    get_documentation_dict,
    get_rules,
)
from sweepai.config.server import (
    DISCORD_FEEDBACK_WEBHOOK_URL,
    ENV,
    GITHUB_LABEL_NAME,
    IS_SELF_HOSTED,
    MONGODB_URI,
    PROGRESS_BASE_URL,
)
from sweepai.core.entities import (
    AssistantRaisedException,
    FileChangeRequest,
    MaxTokensExceeded,
    Message,
    NoFilesException,
    ProposedIssue,
    PullRequest,
    SandboxResponse,
)
from sweepai.core.entities import create_error_logs as entities_create_error_logs
from sweepai.core.pr_reader import PRReader
from sweepai.core.sweep_bot import SweepBot
from sweepai.handlers.create_pr import (
    create_config_pr,
    create_pr_changes,
    safe_delete_sweep_branch,
)
from sweepai.handlers.on_check_suite import clean_gh_logs
from sweepai.utils.buttons import Button, ButtonList, create_action_buttons
from sweepai.utils.chat_logger import ChatLogger
from sweepai.utils.diff import generate_diff
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import (
    CURRENT_USERNAME,
    ClonedRepo,
    convert_pr_draft_field,
    get_github_client,
)
from sweepai.utils.progress import (
    AssistantConversation,
    PaymentContext,
    TicketContext,
    TicketProgress,
    TicketProgressStatus,
)
from sweepai.utils.prompt_constructor import HumanMessagePrompt
from sweepai.utils.str_utils import (
    BOT_SUFFIX,
    FASTER_MODEL_MESSAGE,
    UPDATES_MESSAGE,
    blockquote,
    bot_suffix,
    checkbox_template,
    clean_logs,
    collapsible_template,
    create_checkbox,
    create_collapsible,
    discord_suffix,
    format_sandbox_success,
    get_hash,
    sep,
    stars_suffix,
    strip_sweep,
    to_branch_name,
)
from sweepai.utils.ticket_utils import (
    center,
    fetch_relevant_files,
    fire_and_forget_wrapper,
    log_error,
)
from sweepai.utils.user_settings import UserSettings

# from sandbox.sandbox_utils import Sandbox


sweeping_gif = """<a href="https://github.com/sweepai/sweep"><img class="swing" src="https://raw.githubusercontent.com/sweepai/sweep/main/.assets/sweeping.gif" width="100" style="width:50px; margin-bottom:10px" alt="Sweeping"></a>"""


custom_config = """
extends: relaxed

rules:
    line-length: disable
    indentation: disable
"""

INSTRUCTIONS_FOR_REVIEW = """\
### 💡 To get Sweep to edit this pull request, you can:
* Comment below, and Sweep can edit the entire PR
* Comment on a file, Sweep will only modify the commented file
* Edit the original issue to get Sweep to recreate the PR from scratch"""

email_template = """Hey {name},
<br/><br/>
🚀 I just finished creating a pull request for your issue ({repo_full_name}#{issue_number}) at <a href="{pr_url}">{repo_full_name}#{pr_number}</a>!

<br/><br/>
You can view how I created this pull request <a href="{progress_url}">here</a>.

<h2>Summary</h2>
<blockquote>
{summary}
</blockquote>

<h2>Files Changed</h2>
<ul>
{files_changed}
</ul>

{sweeping_gif}
<br/>
Cheers,
<br/>
Sweep
<br/>"""

FAILING_GITHUB_ACTION_PROMPT = """
The following Github Actions failed on a previous attempt at fixing this issue.
Review the provided logs to ensure that any code modifications you make do not cause these actions to fail again.
{github_action_log}
"""


# Add :eyes: emoji to ticket
def add_emoji(issue: Issue, comment_id: int = None, reaction_content="eyes"):
    item_to_react_to = issue.get_comment(comment_id) if comment_id else issue
    item_to_react_to.create_reaction("eyes")


# If SWEEP_BOT reacted to item_to_react_to with "rocket", then remove it.
def remove_emoji(issue: Issue, comment_id: int = None, content_to_delete="eyes"):
    item_to_react_to = issue.get_comment(comment_id) if comment_id else issue
    reactions = item_to_react_to.get_reactions()
    for reaction in reactions:
        if (
            reaction.content == content_to_delete
            and reaction.user.login == CURRENT_USERNAME
        ):
            item_to_react_to.delete_reaction(reaction.id)


def create_error_logs(
    commit_url_display: str,
    sandbox_response: SandboxResponse,
    status: str = "✓",
):
    return (
        (
            "<br/>"
            + create_collapsible(
                f"Sandbox logs for {commit_url_display} {status}",
                blockquote(
                    "\n\n".join(
                        [
                            create_collapsible(
                                f"<code>{output}</code> {i + 1}/{len(sandbox_response.outputs)} {format_sandbox_success(sandbox_response.success)}",
                                f"<pre>{clean_logs(output)}</pre>",
                                i == len(sandbox_response.outputs) - 1,
                            )
                            for i, output in enumerate(sandbox_response.outputs)
                            if len(sandbox_response.outputs) > 0
                        ]
                    )
                ),
                opened=True,
            )
        )
        if sandbox_response
        else ""
    )


# takes in a list of workflow runs and returns a list of messages containing the logs of the failing runs
def get_failing_gha_logs(runs) -> list[Message]:
    messages: list[Message] = []
    for run in runs:
        # jobs_url
        jobs_url = run.jobs_url
        jobs_response = requests.get(
            jobs_url,
            headers={
                "Accept": "application/vnd.github+json",
                "X-Github-Api-Version": "2022-11-28",
                "Authorization": "Bearer " + os.environ["GITHUB_PAT"],
            },
        )
        if jobs_response.status_code == 200:
            failed_jobs = []
            jobs = jobs_response.json()["jobs"]
            for job in jobs:
                if job["conclusion"] == "failure":
                    failed_jobs.append(job)

            failed_jobs_name_list = []
            for job in failed_jobs:
                # add failed steps
                for step in job["steps"]:
                    if step["conclusion"] == "failure":
                        failed_jobs_name_list.append(
                            f"{job['name']}/{step['number']}_{step['name']}"
                        )
        else:
            logger.error(
                "Failed to get jobs for failing github actions, possible a credentials issue"
            )
            return messages
        # logs url
        logs_url = run.logs_url
        logs_response = requests.get(
            logs_url,
            headers={
                "Accept": "application/vnd.github+json",
                "X-Github-Api-Version": "2022-11-28",
                "Authorization": "Bearer " + os.environ["GITHUB_PAT"],
            },
            allow_redirects=True,
        )
        # Check if the request was successful
        if logs_response.status_code == 200:
            zip_data = io.BytesIO(logs_response.content)
            zip_file = zipfile.ZipFile(zip_data, "r")
            zip_file_names = zip_file.namelist()
            for file in failed_jobs_name_list:
                if f"{file}.txt" in zip_file_names:
                    logs = zip_file.read(f"{file}.txt").decode("utf-8")
                    cleaned_logs, user_message = clean_gh_logs(logs)
                    messages.append(
                        Message(
                            role="user",
                            content=FAILING_GITHUB_ACTION_PROMPT.replace(
                                "{github_action_log}", cleaned_logs
                            ),
                        )
                    )
        else:
            logger.error(
                "Failed to get logs for failing github actions, likely a credentials issue"
            )
            return messages
    return messages


def delete_old_prs(repo: Repository, issue_number: int):
    logger.info("Deleting old PRs...")
    prs = repo.get_pulls(
        state="open",
        sort="created",
        direction="desc",
        base=SweepConfig.get_branch(repo),
    )
    for pr in tqdm(prs.get_page(0)):
        # # Check if this issue is mentioned in the PR, and pr is owned by bot
        # # This is done in create_pr, (pr_description = ...)
        if pr.user.login == CURRENT_USERNAME and f"Fixes #{issue_number}.\n" in pr.body:
            safe_delete_sweep_branch(pr, repo)
            break


def get_comment_header(
    index: int,
    g: Github,
    repo_full_name: str,
    user_settings: UserSettings,
    progress_headers: list[None | str],
    tracking_id: str | None,
    payment_message_start: str,
    user_settings_message: str,
    errored: bool = False,
    pr_message: str = "",
    done: bool = False,
    initial_sandbox_response: int | SandboxResponse = -1,
    initial_sandbox_response_file=None,
    config_pr_url: str | None = None,
):
    config_pr_message = (
        "\n"
        + f"<div align='center'>Install Sweep Configs: <a href='{config_pr_url}'>Pull Request</a></div>"
        if config_pr_url is not None
        else ""
    )
    actions_message = create_action_buttons(
        [
            RESTART_SWEEP_BUTTON,
        ]
    )

    sandbox_execution_message = "\n\n## GitHub Actions failed\n\nThe sandbox appears to be unavailable or down.\n\n"

    if initial_sandbox_response == -1:
        sandbox_execution_message = ""
    elif initial_sandbox_response is not None:
        repo = g.get_repo(repo_full_name)
        commit_hash = repo.get_commits()[0].sha
        success = initial_sandbox_response.outputs and initial_sandbox_response.success
        status = "✓" if success else "X"
        sandbox_execution_message = (
            "\n\n## GitHub Actions"
            + status
            + "\n\nHere are the GitHub Actions logs prior to making any changes:\n\n"
        )
        sandbox_execution_message += entities_create_error_logs(
            f'<a href="https://github.com/{repo_full_name}/commit/{commit_hash}"><code>{commit_hash[:7]}</code></a>',
            initial_sandbox_response,
            initial_sandbox_response_file,
        )
        if success:
            sandbox_execution_message += f"\n\nSandbox passed on the latest `{repo.default_branch}`, so sandbox checks will be enabled for this issue."
        else:
            sandbox_execution_message += "\n\nSandbox failed, so all sandbox checks will be disabled for this issue."

    if index < 0:
        index = 0
    if index == 4:
        return (
            pr_message
            + config_pr_message
            + f"\n\n---\n{user_settings.get_message(completed=True)}"
            + f"\n\n---\n{actions_message}"
            + sandbox_execution_message
        )

    total = len(progress_headers)
    index += 1 if done else 0
    index *= 100 / total
    index = int(index)
    index = min(100, index)
    if errored:
        pbar = f"\n\n<img src='https://progress-bar.dev/{index}/?&title=Errored&width=600' alt='{index}%' />"
        return (
            f"{center(sweeping_gif)}<br/>{center(pbar)}\n\n"
            + f"\n\n---\n{actions_message}"
            + sandbox_execution_message
        )
    pbar = f"\n\n<img src='https://progress-bar.dev/{index}/?&title=Progress&width=600' alt='{index}%' />"
    return (
        f"{center(sweeping_gif)}"
        + (
            center(
                f'\n\n<h2>✨ Track Sweep\'s progress on our <a href="{PROGRESS_BASE_URL}/issues/{tracking_id}">progress dashboard</a>!</h2>'
            )
            if MONGODB_URI is not None
            else ""
        )
        + f"<br/>{center(pbar)}"
        + ("\n" + stars_suffix if index != -1 else "")
        + "\n"
        + center(payment_message_start)
        + f"\n\n---\n{user_settings_message}"
        + config_pr_message
        + f"\n\n---\n{actions_message}"
        + sandbox_execution_message
    )


def on_ticket(
    title: str,
    summary: str,
    issue_number: int,
    issue_url: str,
    username: str,
    repo_full_name: str,
    repo_description: str,
    installation_id: int,
    comment_id: int = None,
    edited: bool = False,
    tracking_id: str | None = None,
):
    with logger.contextualize(
        tracking_id=tracking_id,
    ):
        # we want to pass in the failing github action messages to the next run in order to fix them
        failing_gha_messages: list[Message] = []
        # we rerun this logic 3 times at most if the github actions associated with the created pr fails
        max_pr_attempts = 1
        for run_attempt in range(max_pr_attempts):
            if tracking_id is None:
                tracking_id = get_hash()
            on_ticket_start_time = time()
            logger.info(f"Starting on_ticket with title {title} and summary {summary}")
            (
                title,
                slow_mode,
                do_map,
                subissues_mode,
                sandbox_mode,
                fast_mode,
                lint_mode,
            ) = strip_sweep(title)

            summary = summary or ""
            summary = re.sub(
                "<details (open)?>(\r)?\n<summary>Checklist</summary>.*",
                "",
                summary,
                flags=re.DOTALL,
            ).strip()
            summary = re.sub(
                "---\s+Checklist:(\r)?\n(\r)?\n- \[[ X]\].*",
                "",
                summary,
                flags=re.DOTALL,
            ).strip()
            summary = re.sub(
                "### Details\n\n_No response_", "", summary, flags=re.DOTALL
            )
            summary = re.sub("\n\n", "\n", summary, flags=re.DOTALL)

            repo_name = repo_full_name
            user_token, g = get_github_client(installation_id)
            repo = g.get_repo(repo_full_name)
            current_issue: Issue = repo.get_issue(number=issue_number)
            assignee = current_issue.assignee.login if current_issue.assignee else None
            if assignee is None:
                assignee = current_issue.user.login

            ticket_progress = TicketProgress(
                tracking_id=tracking_id,
                username=username,
                context=TicketContext(
                    title=title,
                    description=summary,
                    repo_full_name=repo_full_name,
                    issue_number=issue_number,
                    is_public=repo.private is False,
                    start_time=int(time()),
                ),
            )
            branch_match = re.search(
                r"([B|b]ranch:) *(?P<branch_name>.+?)(\n|$)", summary
            )
            overrided_branch_name = None
            if branch_match and "branch_name" in branch_match.groupdict():
                overrided_branch_name = (
                    branch_match.groupdict()["branch_name"].strip().strip("`\"'")
                )
                if overrided_branch_name == "_No response_":
                    continue
                SweepConfig.get_branch(repo, overrided_branch_name)

            chat_logger = (
                ChatLogger(
                    {
                        "repo_name": repo_name,
                        "title": title,
                        "summary": summary,
                        "issue_number": issue_number,
                        "issue_url": issue_url,
                        "username": (
                            username if not username.startswith("sweep") else assignee
                        ),
                        "repo_full_name": repo_full_name,
                        "repo_description": repo_description,
                        "installation_id": installation_id,
                        "type": "ticket",
                        "mode": ENV,
                        "comment_id": comment_id,
                        "edited": edited,
                        "tracking_id": tracking_id,
                    },
                    active=True,
                )
                if MONGODB_URI
                else None
            )

            if chat_logger:
                is_paying_user = chat_logger.is_paying_user()
                is_consumer_tier = chat_logger.is_consumer_tier()
                use_faster_model = chat_logger.use_faster_model()
            else:
                is_paying_user = True
                is_consumer_tier = False
                use_faster_model = False

            if use_faster_model:
                raise Exception(FASTER_MODEL_MESSAGE)

            if fast_mode:
                use_faster_model = True

            if not comment_id and not edited and chat_logger and not sandbox_mode:
                fire_and_forget_wrapper(chat_logger.add_successful_ticket)(
                    gpt3=use_faster_model
                )

            organization, repo_name = repo_full_name.split("/")
            metadata = {
                "issue_url": issue_url,
                "repo_full_name": repo_full_name,
                "organization": organization,
                "repo_name": repo_name,
                "repo_description": repo_description,
                "username": username,
                "comment_id": comment_id,
                "title": title,
                "installation_id": installation_id,
                "function": "on_ticket",
                "edited": edited,
                "model": "gpt-3.5" if use_faster_model else "gpt-4",
                "tier": "pro" if is_paying_user else "free",
                "mode": ENV,
                "slow_mode": slow_mode,
                "do_map": do_map,
                "subissues_mode": subissues_mode,
                "sandbox_mode": sandbox_mode,
                "fast_mode": fast_mode,
                "is_self_hosted": IS_SELF_HOSTED,
                "tracking_id": tracking_id,
            }
            if run_attempt == 0:
                # we want to capture restart and start events separately in posthog
                fire_and_forget_wrapper(posthog.capture)(
                    username, "started", properties=metadata
                )
            else:
                fire_and_forget_wrapper(posthog.capture)(
                    username, "on_ticket_restarted", properties=metadata
                )

            try:
                if current_issue.state == "closed":
                    fire_and_forget_wrapper(posthog.capture)(
                        username,
                        "issue_closed",
                        properties={
                            **metadata,
                            "duration": round(time() - on_ticket_start_time),
                        },
                    )
                    return {"success": False, "reason": "Issue is closed"}

                fire_and_forget_wrapper(add_emoji)(current_issue, comment_id)
                fire_and_forget_wrapper(remove_emoji)(
                    current_issue, comment_id, content_to_delete="rocket"
                )
                fire_and_forget_wrapper(current_issue.edit)(body=summary)

                replies_text = ""
                summary = summary if summary else ""

                fire_and_forget_wrapper(delete_old_prs)(repo, issue_number)

                if not sandbox_mode:
                    progress_headers = [
                        None,
                        "Step 1: 🔎 Searching",
                        "Step 2: ⌨️ Coding",
                        "Step 3: 🔁 Code Review",
                    ]
                else:
                    progress_headers = [
                        None,
                        "📖 Reading File",
                        "🛠️ Executing Sandbox",
                    ]

                issue_comment = None
                payment_message, payment_message_start = get_payment_messages(
                    chat_logger
                )

                ticket_progress.context.payment_context = PaymentContext(
                    use_faster_model=use_faster_model,
                    pro_user=is_paying_user,
                    daily_tickets_used=(
                        chat_logger.get_ticket_count(use_date=True)
                        if chat_logger
                        else 0
                    ),
                    monthly_tickets_used=(
                        chat_logger.get_ticket_count() if chat_logger else 0
                    ),
                )
                ticket_progress.save()

                config_pr_url = None
                user_settings = UserSettings.from_username(username=username)
                user_settings_message = user_settings.get_message()

                cloned_repo = ClonedRepo(
                    repo_full_name,
                    installation_id=installation_id,
                    token=user_token,
                    repo=repo,
                    branch=overrided_branch_name,
                )
                # check that repo's directory is non-empty
                if os.listdir(cloned_repo.cached_dir) == []:
                    logger.info("Empty repo")
                    first_comment = (
                        "Sweep is currently not supported on empty repositories. Please add some"
                        f" code to your repository and try again.\n{sep}##"
                        f" {progress_headers[1]}\n{bot_suffix}{discord_suffix}"
                    )
                    if issue_comment is None:
                        issue_comment = current_issue.create_comment(
                            first_comment + BOT_SUFFIX
                        )
                    else:
                        issue_comment.edit(first_comment + BOT_SUFFIX)
                    return {"success": False}
                indexing_message = (
                    "I'm searching for relevant snippets in your repository. If this is your first"
                    " time using Sweep, I'm indexing your repository. You can monitor the progress using the progress dashboard"
                )
                first_comment = (
                    f"{get_comment_header(0, g, repo_full_name, user_settings, progress_headers, tracking_id, payment_message_start, user_settings_message)}\n{sep}I am currently looking into this ticket! I"
                    " will update the progress of the ticket in this comment. I am currently"
                    f" searching through your code, looking for relevant snippets.\n{sep}##"
                    f" {progress_headers[1]}\n{indexing_message}{bot_suffix}{discord_suffix}"
                )
                # Find Sweep's previous comment
                comments = []
                for comment in current_issue.get_comments():
                    comments.append(comment)
                    if comment.user.login == CURRENT_USERNAME:
                        issue_comment = comment
                        break
                if issue_comment is None:
                    issue_comment = current_issue.create_comment(first_comment)
                else:
                    fire_and_forget_wrapper(issue_comment.edit)(first_comment)
                old_edit = issue_comment.edit
                issue_comment.edit = lambda msg: old_edit(msg + BOT_SUFFIX)
                past_messages = {}
                current_index = 0
                table = None
                initial_sandbox_response = -1
                initial_sandbox_response_file = None

                def refresh_token():
                    user_token, g = get_github_client(installation_id)
                    repo = g.get_repo(repo_full_name)
                    return user_token, g, repo

                def edit_sweep_comment(
                    message: str,
                    index: int,
                    pr_message="",
                    done=False,
                    add_bonus_message=True,
                ):
                    nonlocal current_index, user_token, g, repo, issue_comment, initial_sandbox_response, initial_sandbox_response_file
                    # -1 = error, -2 = retry
                    # Only update the progress bar if the issue generation errors.
                    errored = index == -1
                    if index >= 0:
                        past_messages[index] = message
                        current_index = index

                    agg_message = None
                    # Include progress history
                    # index = -2 is reserved for
                    for i in range(
                        current_index + 2
                    ):  # go to next header (for Working on it... text)
                        if i == 0 or i >= len(progress_headers):
                            continue  # skip None header
                        header = progress_headers[i]
                        if header is not None:
                            header = "## " + header + "\n"
                        else:
                            header = "No header\n"
                        msg = header + (past_messages.get(i) or "Working on it...")
                        if agg_message is None:
                            agg_message = msg
                        else:
                            agg_message = agg_message + f"\n{sep}" + msg

                    suffix = bot_suffix + discord_suffix
                    if errored:
                        agg_message = (
                            "## ❌ Unable to Complete PR"
                            + "\n"
                            + message
                            + (
                                "\n\nFor bonus GPT-4 tickets, please report this bug on"
                                f" **[Discord](https://discord.gg/invite/sweep)** (tracking ID: `{tracking_id}`)."
                                if add_bonus_message
                                else ""
                            )
                        )
                        if table is not None:
                            agg_message = (
                                agg_message
                                + f"\n{sep}Please look at the generated plan. If something looks"
                                f" wrong, please add more details to your issue.\n\n{table}"
                            )
                        suffix = bot_suffix  # don't include discord suffix for error messages

                    # Update the issue comment
                    msg = f"{get_comment_header(current_index, g, repo_full_name, user_settings, progress_headers, tracking_id, payment_message_start, user_settings_message, errored=errored, pr_message=pr_message, done=done, initial_sandbox_response=initial_sandbox_response, initial_sandbox_response_file=initial_sandbox_response_file, config_pr_url=config_pr_url)}\n{sep}{agg_message}{suffix}"
                    try:
                        issue_comment.edit(msg)
                    except BadCredentialsException:
                        logger.error(
                            f"Bad credentials, refreshing token (tracking ID: `{tracking_id}`)"
                        )
                        user_token, g = get_github_client(installation_id)
                        repo = g.get_repo(repo_full_name)

                        issue_comment = None
                        for comment in comments:
                            if comment.user.login == CURRENT_USERNAME:
                                issue_comment = comment
                        current_issue = repo.get_issue(number=issue_number)
                        if issue_comment is None:
                            issue_comment = current_issue.create_comment(msg)
                        else:
                            issue_comment = [
                                comment
                                for comment in current_issue.get_comments()
                                if comment.user.login == CURRENT_USERNAME
                            ][0]
                            issue_comment.edit(msg)

                if use_faster_model:
                    edit_sweep_comment(
                        FASTER_MODEL_MESSAGE, -1, add_bonus_message=False
                    )
                    posthog.capture(
                        username,
                        "ran_out_of_tickets",
                        properties={
                            **metadata,
                            "duration": round(time() - on_ticket_start_time),
                        },
                    )
                    return {
                        "success": False,
                        "error_message": "We deprecated supporting GPT 3.5.",
                    }

                if sandbox_mode:
                    handle_sandbox_mode(
                        title, repo_full_name, repo, ticket_progress, edit_sweep_comment
                    )
                    return {"success": True}

                if len(title + summary) < 20:
                    logger.info("Issue too short")
                    edit_sweep_comment(
                        (
                            f"Please add more details to your issue. I need at least 20 characters"
                            f" to generate a plan. Please join our Discord server for support (tracking_id={tracking_id})"
                        ),
                        -1,
                    )
                    posthog.capture(
                        username,
                        "issue_too_short",
                        properties={
                            **metadata,
                            "duration": round(time() - on_ticket_start_time),
                        },
                    )
                    return {"success": True}

                prs_extracted = PRReader.extract_prs(repo, summary)
                message_summary = summary
                if prs_extracted:
                    message_summary += "\n\n" + prs_extracted
                    edit_sweep_comment(
                        create_collapsible(
                            "I found that you mentioned the following Pull Requests that might be important:",
                            blockquote(
                                prs_extracted,
                            ),
                        ),
                        1,
                    )

                try:
                    # search/context manager
                    logger.info("Searching for relevant snippets...")
                    snippets, tree, _ = fetch_relevant_files(
                        cloned_repo,
                        title,
                        message_summary,
                        replies_text,
                        username,
                        metadata,
                        on_ticket_start_time,
                        tracking_id,
                        is_paying_user,
                        is_consumer_tier,
                        issue_url,
                        chat_logger,
                        ticket_progress,
                    )
                except Exception:
                    edit_sweep_comment(
                        (
                            "It looks like an issue has occurred around fetching the files."
                            " Perhaps the repo failed to initialized. If this error persists"
                            f" contact team@sweep.dev.\n\n> @{username}, editing this issue description to include more details will automatically make me relaunch. Please join our Discord server for support (tracking_id={tracking_id})"
                        ),
                        -1,
                    )
                    raise Exception("Failed to fetch files")
                _user_token, g = get_github_client(installation_id)
                user_token, g, repo = refresh_token()
                cloned_repo.token = user_token
                repo = g.get_repo(repo_full_name)
                ticket_progress.search_progress.indexing_progress = (
                    ticket_progress.search_progress.indexing_total
                )
                ticket_progress.status = TicketProgressStatus.PLANNING
                ticket_progress.save()

                # Fetch git commit history
                if not repo_description:
                    repo_description = "No description provided."

                message_summary += replies_text
                # removed external search as it provides no real value and only adds noise
                # external_results = ExternalSearcher.extract_summaries(message_summary)
                # if external_results:
                #     message_summary += "\n\n" + external_results

                get_documentation_dict(repo)
                docs_results = ""
                human_message = HumanMessagePrompt(
                    repo_name=repo_name,
                    issue_url=issue_url,
                    username=username,
                    repo_description=repo_description.strip(),
                    title=title,
                    summary=message_summary,
                    snippets=snippets,
                    tree=tree,
                )

                sweep_bot = SweepBot.from_system_message_content(
                    human_message=human_message,
                    repo=repo,
                    is_reply=bool(comments),
                    chat_logger=chat_logger,
                    cloned_repo=cloned_repo,
                    ticket_progress=ticket_progress,
                )

                # Check repository for sweep.yml file.
                sweep_yml_exists = False
                sweep_yml_failed = False
                for content_file in repo.get_contents(""):
                    if content_file.name == "sweep.yaml":
                        sweep_yml_exists = True

                        # Check if YAML is valid
                        yaml_content = content_file.decoded_content.decode("utf-8")
                        sweep_yaml_dict = {}
                        try:
                            sweep_yaml_dict = yaml.safe_load(yaml_content)
                        except Exception:
                            logger.error(f"Failed to load YAML file: {yaml_content}")
                        if len(sweep_yaml_dict) > 0:
                            break
                        linter_config = yamllint_config.YamlLintConfig(custom_config)
                        problems = list(linter.run(yaml_content, linter_config))
                        if problems:
                            errors = [
                                f"Line {problem.line}: {problem.desc} (rule: {problem.rule})"
                                for problem in problems
                            ]
                            error_message = "\n".join(errors)
                            markdown_error_message = f"**There is something wrong with your [sweep.yaml](https://github.com/{repo_full_name}/blob/main/sweep.yaml):**\n```\n{error_message}\n```"
                            sweep_yml_failed = True
                            logger.error(markdown_error_message)
                            edit_sweep_comment(markdown_error_message, -1)
                        else:
                            logger.info("The YAML file is valid. No errors found.")
                        break

                # If sweep.yaml does not exist, then create a new PR that simply creates the sweep.yaml file.
                if not sweep_yml_exists:
                    try:
                        logger.info("Creating sweep.yaml file...")
                        config_pr = create_config_pr(sweep_bot, cloned_repo=cloned_repo)
                        config_pr_url = config_pr.html_url
                        edit_sweep_comment(message="", index=-2)
                    except SystemExit:
                        raise SystemExit
                    except Exception as e:
                        logger.error(
                            "Failed to create new branch for sweep.yaml file.\n",
                            e,
                            traceback.format_exc(),
                        )
                else:
                    logger.info("sweep.yaml file already exists.")

                try:
                    # ANALYZE SNIPPETS
                    newline = "\n"
                    edit_sweep_comment(
                        "I found the following snippets in your repository. I will now analyze"
                        " these snippets and come up with a plan."
                        + "\n\n"
                        + create_collapsible(
                            "Some code snippets I think are relevant in decreasing order of relevance (click to expand). If some file is missing from here, you can mention the path in the ticket description.",
                            "\n".join(
                                [
                                    f"https://github.com/{organization}/{repo_name}/blob/{repo.get_commits()[0].sha}/{snippet.file_path}#L{max(snippet.start, 1)}-L{min(snippet.end, snippet.content.count(newline) - 1)}\n"
                                    for snippet in snippets
                                ]
                            ),
                        )
                        + (
                            create_collapsible(
                                "I also found that you mentioned the following Pull Requests that may be helpful:",
                                blockquote(prs_extracted),
                            )
                            if prs_extracted
                            else ""
                        )
                        # removed external results as it provides no real value and only adds noise
                        # + (
                        #     create_collapsible(
                        #         "I also found the following external resources that might be helpful:",
                        #         f"\n\n{external_results}\n\n",
                        #     )
                        #     if external_results
                        #     else ""
                        # )
                        + (f"\n\n{docs_results}\n\n" if docs_results else ""),
                        1,
                    )

                    if do_map:
                        subissues: list[ProposedIssue] = sweep_bot.generate_subissues()
                        edit_sweep_comment(
                            "I'm creating the following subissues:\n\n"
                            + "\n\n".join(
                                [
                                    f"#{subissue.title}:\n" + blockquote(subissue.body)
                                    for subissue in subissues
                                ]
                            ),
                            2,
                        )
                        for subissue in tqdm(subissues):
                            subissue.issue_id = repo.create_issue(
                                title="Sweep: " + subissue.title,
                                body=subissue.body
                                + f"\n\nParent issue: #{issue_number}",
                                assignee=username,
                            ).number
                        subissues_checklist = "\n\n".join(
                            [
                                f"- [ ] #{subissue.issue_id}\n\n"
                                + blockquote(f"**{subissue.title}**\n{subissue.body}")
                                for subissue in subissues
                            ]
                        )
                        current_issue.edit(
                            body=summary
                            + "\n\n---\n\nChecklist:\n\n"
                            + subissues_checklist
                        )
                        edit_sweep_comment(
                            "I finished creating the subissues! Track them at:\n\n"
                            + "\n".join(
                                f"* #{subissue.issue_id}" for subissue in subissues
                            ),
                            3,
                            done=True,
                        )
                        edit_sweep_comment("N/A", 4)
                        edit_sweep_comment("I finished creating all the subissues.", 5)
                        posthog.capture(
                            username,
                            "subissues_created",
                            properties={
                                **metadata,
                                "count": len(subissues),
                                "duration": round(time() - on_ticket_start_time),
                            },
                        )
                        return {"success": True}

                    logger.info("Fetching files to modify/create...")
                    non_python_count = sum(
                        not file_path.endswith(".py")
                        and not file_path.endswith(".ipynb")
                        and not file_path.endswith(".md")
                        for file_path in human_message.get_file_paths()
                    )
                    python_count = (
                        len(human_message.get_file_paths()) - non_python_count
                    )
                    is_python_issue = (
                        python_count >= non_python_count and python_count > 0
                    )
                    posthog.capture(
                        username,
                        "is_python_issue",
                        properties={"is_python_issue": is_python_issue},
                    )
                    file_change_requests, plan = sweep_bot.get_files_to_change(
                        is_python_issue
                    )
                    ticket_progress.planning_progress.file_change_requests = (
                        file_change_requests
                    )
                    ticket_progress.coding_progress.file_change_requests = (
                        file_change_requests
                    )
                    ticket_progress.coding_progress.assistant_conversations = [
                        AssistantConversation() for fcr in file_change_requests
                    ]
                    ticket_progress.status = TicketProgressStatus.CODING
                    ticket_progress.save()

                    if not file_change_requests:
                        if len(title + summary) < 60:
                            edit_sweep_comment(
                                (
                                    "Sorry, I could not find any files to modify, can you please"
                                    " provide more details? Please make sure that the title and"
                                    " summary of the issue are at least 60 characters."
                                ),
                                -1,
                            )
                        else:
                            edit_sweep_comment(
                                (
                                    "Sorry, I could not find any files to modify, can you please"
                                    " provide more details?"
                                ),
                                -1,
                            )
                        raise Exception("No files to modify.")

                    file_change_requests: list[
                        FileChangeRequest
                    ] = sweep_bot.validate_file_change_requests(
                        file_change_requests,
                    )
                    ticket_progress.planning_progress.file_change_requests = (
                        file_change_requests
                    )
                    ticket_progress.coding_progress.assistant_conversations = [
                        AssistantConversation() for fcr in file_change_requests
                    ]
                    ticket_progress.save()

                    table = tabulate(
                        [
                            [
                                file_change_request.entity_display,
                                file_change_request.instructions_display.replace(
                                    "\n", "<br/>"
                                ).replace("```", "\\```"),
                            ]
                            for file_change_request in file_change_requests
                            if file_change_request.change_type != "check"
                        ],
                        headers=["File Path", "Proposed Changes"],
                        tablefmt="pipe",
                    )

                    logger.info("Generating PR...")
                    pull_request = PullRequest(
                        title="Sweep: " + title,
                        branch_name="sweep/" + to_branch_name(title),
                        content="",
                    )
                    logger.info("Making PR...")

                    ticket_progress.context.branch_name = pull_request.branch_name
                    ticket_progress.save()

                    files_progress: list[tuple[str, str, str, str]] = [
                        (
                            file_change_request.entity_display,
                            file_change_request.instructions_display,
                            "⏳ In Progress",
                            "",
                        )
                        for file_change_request in file_change_requests
                    ]

                    checkboxes_progress: list[tuple[str, str, str]] = [
                        (
                            file_change_request.entity_display,
                            file_change_request.instructions_display,
                            " ",
                        )
                        for file_change_request in file_change_requests
                        if not file_change_request.change_type == "check"
                    ]
                    checkboxes_contents = "\n".join(
                        [
                            create_checkbox(
                                f"`{filename}`", blockquote(instructions), check == "X"
                            )
                            for filename, instructions, check in checkboxes_progress
                        ]
                    )
                    create_collapsible("Checklist", checkboxes_contents, opened=True)

                    file_change_requests[0].status = "running"

                    condensed_checkboxes_contents = "\n".join(
                        [
                            create_checkbox(f"`{filename}`", "", check == "X").strip()
                            for filename, instructions, check in checkboxes_progress
                        ]
                    )
                    condensed_checkboxes_collapsible = create_collapsible(
                        "Checklist", condensed_checkboxes_contents, opened=True
                    )

                    current_issue = repo.get_issue(number=issue_number)
                    current_issue.edit(
                        body=summary + "\n\n" + condensed_checkboxes_collapsible
                    )

                    delete_branch = False

                    # this is to prevent failing_gha_messages from being modified by the generator
                    additional_messages = deepcopy(failing_gha_messages)

                    generator = create_pr_changes(
                        file_change_requests,
                        pull_request,
                        sweep_bot,
                        username,
                        installation_id,
                        issue_number,
                        chat_logger=chat_logger,
                        base_branch=overrided_branch_name,
                        additional_messages=additional_messages,
                    )
                    edit_sweep_comment(checkboxes_contents, 2)
                    if not file_change_requests:
                        raise NoFilesException()
                    response = {
                        "error": Exception(
                            "Sweep failed to make code changes! This could mean that GPT-4 failed to find the correct lines of code to modify or that GPT-4 did not respond in our specified format. Sometimes, retrying will fix this error. Otherwise, reach out to our Discord server for support (tracking_id={tracking_id})."
                        )
                    }

                    changed_files = []
                    for item in generator:
                        if isinstance(item, dict):
                            response = item
                            break
                        (
                            file_change_request,
                            changed_file,
                            sandbox_response,
                            commit,
                            file_change_requests,
                        ) = item
                        changed_files.append(file_change_request.filename)
                        sandbox_response: SandboxResponse | None = sandbox_response
                        logger.info(sandbox_response)
                        commit_hash: str = (
                            commit
                            if isinstance(commit, str)
                            else (
                                commit.sha
                                if commit is not None
                                else repo.get_branch(
                                    pull_request.branch_name
                                ).commit.sha
                            )
                        )
                        commit_url = (
                            f"https://github.com/{repo_full_name}/commit/{commit_hash}"
                        )
                        commit_url_display = (
                            f"<a href='{commit_url}'><code>{commit_hash[:7]}</code></a>"
                        )
                        create_error_logs(
                            commit_url_display,
                            sandbox_response,
                            status=(
                                "✓"
                                if (
                                    sandbox_response is None or sandbox_response.success
                                )
                                else "❌"
                            ),
                        )
                        checkboxes_progress = [
                            (
                                file_change_request.display_summary
                                + " "
                                + file_change_request.status_display
                                + " "
                                + (file_change_request.commit_hash_url or "")
                                + f" [Edit]({file_change_request.get_edit_url(repo.full_name, pull_request.branch_name)})",
                                file_change_request.instructions_ticket_display
                                + f"\n\n{file_change_request.diff_display}",
                                (
                                    "X"
                                    if file_change_request.status
                                    in ("succeeded", "failed")
                                    else " "
                                ),
                            )
                            for file_change_request in file_change_requests
                        ]
                        checkboxes_contents = "\n".join(
                            [
                                checkbox_template.format(
                                    check=check,
                                    filename=filename,
                                    instructions=blockquote(instructions),
                                )
                                for filename, instructions, check in checkboxes_progress
                            ]
                        )
                        collapsible_template.format(
                            summary="Checklist",
                            body=checkboxes_contents,
                            opened="open",
                        )
                        condensed_checkboxes_contents = "\n".join(
                            [
                                checkbox_template.format(
                                    check=check,
                                    filename=filename,
                                    instructions="",
                                ).strip()
                                for filename, instructions, check in checkboxes_progress
                                if not instructions.lower().startswith("run")
                            ]
                        )
                        condensed_checkboxes_collapsible = collapsible_template.format(
                            summary="Checklist",
                            body=condensed_checkboxes_contents,
                            opened="open",
                        )

                        try:
                            current_issue = repo.get_issue(number=issue_number)
                        except BadCredentialsException:
                            user_token, g, repo = refresh_token()
                            cloned_repo.token = user_token

                        current_issue.edit(
                            body=summary + "\n\n" + condensed_checkboxes_collapsible
                        )

                        logger.info(files_progress)
                        logger.info(f"Edited {file_change_request.entity_display}")
                        edit_sweep_comment(checkboxes_contents, 2)
                    if not response.get("success"):
                        raise Exception(f"Failed to create PR: {response.get('error')}")

                    checkboxes_contents = "\n".join(
                        [
                            checkbox_template.format(
                                check=check,
                                filename=filename,
                                instructions=blockquote(instructions),
                            )
                            for filename, instructions, check in checkboxes_progress
                        ]
                    )
                    condensed_checkboxes_contents = "\n".join(
                        [
                            checkbox_template.format(
                                check=check,
                                filename=filename,
                                instructions="",
                            ).strip()
                            for filename, instructions, check in checkboxes_progress
                            if not instructions.lower().startswith("run")
                        ]
                    )
                    condensed_checkboxes_collapsible = collapsible_template.format(
                        summary="Checklist",
                        body=condensed_checkboxes_contents,
                        opened="open",
                    )
                    for _ in range(3):
                        try:
                            current_issue.edit(
                                body=summary + "\n\n" + condensed_checkboxes_collapsible
                            )
                            break
                        except Exception:
                            from time import sleep

                            sleep(1)
                    edit_sweep_comment(checkboxes_contents, 2)

                    pr_changes = response["pull_request"]
                    # change the body here
                    diff_text = get_branch_diff_text(
                        repo=repo,
                        branch=pull_request.branch_name,
                        base_branch=overrided_branch_name,
                    )
                    new_description = PRDescriptionBot().describe_diffs(
                        diff_text,
                        pull_request.title,
                    )
                    # TODO: update the title as well
                    if new_description:
                        pr_changes.body = (
                            f"{new_description}\n\nFixes"
                            f" #{issue_number}.\n\n---\n\n{UPDATES_MESSAGE}\n\n---\n\n{INSTRUCTIONS_FOR_REVIEW}"
                        )

                    edit_sweep_comment(
                        "I have finished coding the issue. I am now reviewing it for completeness.",
                        3,
                    )
                    change_location = f" [`{pr_changes.pr_head}`](https://github.com/{repo_full_name}/commits/{pr_changes.pr_head}).\n\n"
                    review_message = (
                        "Here are my self-reviews of my changes at" + change_location
                    )

                    try:
                        fire_and_forget_wrapper(remove_emoji)(content_to_delete="eyes")
                    except SystemExit:
                        raise SystemExit
                    except Exception:
                        pass

                    changes_required, review_message = False, ""
                    if changes_required:
                        edit_sweep_comment(
                            review_message
                            + "\n\nI finished incorporating these changes.",
                            3,
                        )
                    else:
                        edit_sweep_comment(
                            f"I have finished reviewing the code for completeness. I did not find errors for {change_location}",
                            3,
                        )

                    pr_actions_message = (
                        create_action_buttons(
                            [
                                SWEEP_GOOD_FEEDBACK,
                                SWEEP_BAD_FEEDBACK,
                            ],
                            header="### PR Feedback (click)\n",
                        )
                        + "\n"
                        if DISCORD_FEEDBACK_WEBHOOK_URL is not None
                        else ""
                    )
                    revert_buttons = []
                    for changed_file in set(changed_files):
                        revert_buttons.append(
                            Button(label=f"{RESET_FILE} {changed_file}")
                        )
                    revert_buttons_list = ButtonList(
                        buttons=revert_buttons, title=REVERT_CHANGED_FILES_TITLE
                    )

                    rule_buttons = []
                    repo_rules = get_rules(repo) or []
                    if repo_rules != [""] and repo_rules != []:
                        for rule in repo_rules or []:
                            if rule:
                                rule_buttons.append(
                                    Button(label=f"{RULES_LABEL} {rule}")
                                )
                        if len(repo_rules) == 0:
                            for rule in DEFAULT_RULES:
                                rule_buttons.append(
                                    Button(label=f"{RULES_LABEL} {rule}")
                                )

                    rules_buttons_list = ButtonList(
                        buttons=rule_buttons, title=RULES_TITLE
                    )

                    sandbox_passed = None
                    for file_change_request in file_change_requests:
                        if file_change_request.change_type == "check":
                            if (
                                file_change_request.sandbox_response
                                and file_change_request.sandbox_response.error_messages
                            ):
                                sandbox_passed = False
                            elif sandbox_passed is None:
                                sandbox_passed = True

                    # delete failing sweep yaml if applicable
                    if sweep_yml_failed:
                        try:
                            repo.delete_file(
                                "sweep.yaml",
                                "Delete failing sweep.yaml",
                                branch=pr_changes.pr_head,
                                sha=repo.get_contents("sweep.yaml").sha,
                            )
                        except Exception:
                            pass

                    # create draft pr, then convert to regular pr if it all is well
                    is_draft_pr = True
                    # if we are on our last attempt we just create a normal pr
                    if run_attempt >= max_pr_attempts - 1:
                        is_draft_pr = False
                    pr: GithubPullRequest = repo.create_pull(
                        title=pr_changes.title,
                        body=pr_actions_message + pr_changes.body,
                        head=pr_changes.pr_head,
                        base=overrided_branch_name or SweepConfig.get_branch(repo),
                        # TODO: reenable it later
                        draft=is_draft_pr and False,
                    )

                    try:
                        pr.add_to_assignees(username)
                    except Exception as e:
                        logger.error(
                            f"Failed to add assignee {username}: {e}, probably a bot."
                        )

                    ticket_progress.status = TicketProgressStatus.COMPLETE
                    ticket_progress.context.done_time = time()
                    ticket_progress.context.pr_id = pr.number
                    ticket_progress.save()

                    if revert_buttons:
                        pr.create_issue_comment(
                            revert_buttons_list.serialize() + BOT_SUFFIX
                        )
                    if rule_buttons:
                        pr.create_issue_comment(
                            rules_buttons_list.serialize() + BOT_SUFFIX
                        )

                    # add comments before labelling
                    pr.add_to_labels(GITHUB_LABEL_NAME)
                    current_issue.create_reaction("rocket")
                    heres_pr_message = f'<h1 align="center">🚀 Here\'s the PR! <a href="{pr.html_url}">#{pr.number}</a></h1>'
                    progress_message = f'<div align="center"><b>See Sweep\'s progress at <a href="{PROGRESS_BASE_URL}/issues/{tracking_id}">the progress dashboard</a>!</b></div>'
                    edit_sweep_comment(
                        review_message + "\n\nSuccess! 🚀",
                        4,
                        pr_message=(
                            f"{center(heres_pr_message)}\n{center(progress_message)}\n{center(payment_message_start)}"
                        ),
                        done=True,
                    )

                    user_settings = UserSettings.from_username(username=username)
                    user = g.get_user(username)
                    full_name = user.name or user.login
                    name = full_name.split(" ")[0]
                    files_changed = []
                    for fcr in file_change_requests:
                        if fcr.change_type in ("create", "modify"):
                            diff = list(
                                difflib.unified_diff(
                                    (fcr.old_content or "").splitlines() or [],
                                    (fcr.new_content or "").splitlines() or [],
                                    lineterm="",
                                )
                            )
                            added = sum(
                                1
                                for line in diff
                                if line.startswith("+") and not line.startswith("+++")
                            )
                            removed = sum(
                                1
                                for line in diff
                                if line.startswith("-") and not line.startswith("---")
                            )
                            files_changed.append(
                                f"<code>{fcr.filename}</code> (+{added}/-{removed})"
                            )
                    user_settings.send_email(
                        subject=f"Sweep Pull Request Complete for {repo_name}#{issue_number} {title}",
                        html=email_template.format(
                            name=name,
                            pr_url=pr.html_url,
                            issue_number=issue_number,
                            repo_full_name=repo_full_name,
                            pr_number=pr.number,
                            progress_url=f"{PROGRESS_BASE_URL}/issues/{tracking_id}",
                            summary=markdown.markdown(pr_changes.body),
                            files_changed="\n".join(
                                [f"<li>{item}</li>" for item in files_changed]
                            ),
                            sweeping_gif=sweeping_gif,
                        ),
                    )

                    # TODO: re-enable this later
                    # poll for github to check when gha are done or not
                    pr_created_successfully = False
                    total_poll_attempts = 0
                    if False:
                        while True:
                            logger.info(
                                f"Polling to see if Github Actions have finished... {total_poll_attempts}"
                            )
                            # we wait at most 60 minutes
                            if total_poll_attempts >= 60:
                                pr_created_successfully = False
                                break
                            else:
                                # wait one minute between check attempts
                                total_poll_attempts += 1
                                from time import sleep

                                sleep(60)
                            runs = list(repo.get_workflow_runs(branch=pr.head.ref))
                            # if all runs have succeeded, break
                            if all([run.conclusion == "success" for run in runs]):
                                pr_created_successfully = True
                                break
                            # if any of them have failed we retry
                            if any([run.conclusion == "failure" for run in runs]):
                                pr_created_successfully = False
                                failed_runs = [
                                    run for run in runs if run.conclusion == "failure"
                                ]

                                failed_gha_logs: list[Message] = get_failing_gha_logs(
                                    failed_runs
                                )
                                if failed_gha_logs:
                                    failing_gha_messages.extend(failed_gha_logs)
                                logger.info(
                                    f"Rerunning issue {issue_url} as some workflows failed! Rerun attempt {run_attempt + 1}"
                                )
                                # clean up by closing pr and deleting branch associated with pr before restarting on_ticket logic
                                # unless this is sweep's last attempt
                                if run_attempt < 2:
                                    try:
                                        pr.edit(
                                            state="closed",
                                            title=pr.title
                                            + f" (Closed due to failing Github Action: Attempt {run_attempt + 1})",
                                        )
                                        if pr.head.ref.startswith("sweep"):
                                            repo.get_git_ref(
                                                f"heads/{pr.head.ref}"
                                            ).delete()
                                    except Exception as e:
                                        logger.error(
                                            f"Failed to clean up branch {pr.head.ref} and pr before restarting: {e}"
                                        )
                                break
                            # if none of the runs have completed we wait and poll github
                            logger.info(
                                "No Github Actions have failed yet and not all have succeeded yet, waiting for 60 seconds before polling again..."
                            )
                        # break from main for loop
                        if pr_created_successfully:
                            logger.info(
                                f"All Github Actions have finished successfully! It took {run_attempt + 1} attempts to create the PR. It took {total_poll_attempts + 1} minutes for all Github Actions to finish."
                            )
                            # convert draft pr to normal one
                            convert_pr_draft_field(pr, is_draft=False)
                            break

                except MaxTokensExceeded as e:
                    logger.info("Max tokens exceeded")
                    ticket_progress.status = TicketProgressStatus.ERROR
                    ticket_progress.error_message = "Max tokens exceeded. Feel free to add more details to the issue descript for Sweep to better address it, or alternatively, reach out to Kevin or William for help at https://discord.gg/sweep."
                    ticket_progress.save()
                    log_error(
                        is_paying_user,
                        is_consumer_tier,
                        username,
                        issue_url,
                        "Max Tokens Exceeded",
                        str(e) + "\n" + traceback.format_exc(),
                        priority=2,
                    )
                    if chat_logger and chat_logger.is_paying_user():
                        edit_sweep_comment(
                            (
                                f"Sorry, I could not edit `{e.filename}` as this file is too long."
                                " We are currently working on improved file streaming to address"
                                " this issue.\n"
                            ),
                            -1,
                        )
                    else:
                        edit_sweep_comment(
                            (
                                f"Sorry, I could not edit `{e.filename}` as this file is too"
                                " long.\n\nIf this file is incorrect, please describe the desired"
                                " file in the prompt. However, if you would like to edit longer"
                                " files, consider upgrading to [Sweep Pro](https://sweep.dev/) for"
                                " longer context lengths.\n"
                            ),
                            -1,
                        )
                    delete_branch = True
                    raise e
                except NoFilesException as e:
                    ticket_progress.status = TicketProgressStatus.ERROR
                    ticket_progress.error_message = "Sweep could not find files to modify to address this issue. Feel free to add more details to the issue descript for Sweep to better address it, or alternatively, reach out to Kevin or William for help at https://discord.gg/sweep."
                    ticket_progress.save()

                    logger.info("Sweep could not find files to modify")
                    log_error(
                        is_paying_user,
                        is_consumer_tier,
                        username,
                        issue_url,
                        "Sweep could not find files to modify",
                        str(e) + "\n" + traceback.format_exc(),
                        priority=2,
                    )
                    edit_sweep_comment(
                        (
                            "Sorry, Sweep could not find any appropriate files to edit to address"
                            " this issue. If this is a mistake, please provide more context and Sweep"
                            f" will retry!\n\n> @{username}, please edit the issue description to"
                            " include more details about this issue."
                        ),
                        -1,
                    )
                    delete_branch = True
                    raise e
                except openai.BadRequestError as e:
                    ticket_progress.status = TicketProgressStatus.ERROR
                    ticket_progress.error_message = "Sorry, it looks like there is an error with communicating with OpenAI. If this error persists, reach out to Kevin or William for help at https://discord.gg/sweep."
                    ticket_progress.save()

                    logger.error(traceback.format_exc())
                    logger.error(e)
                    edit_sweep_comment(
                        (
                            "I'm sorry, but it looks our model has ran out of context length. We're"
                            " trying to make this happen less, but one way to mitigate this is to"
                            " code smaller files. If this error persists report it at"
                            " https://discord.gg/sweep."
                        ),
                        -1,
                    )
                    log_error(
                        is_paying_user,
                        is_consumer_tier,
                        username,
                        issue_url,
                        "Context Length",
                        str(e) + "\n" + traceback.format_exc(),
                        priority=2,
                    )
                    posthog.capture(
                        username,
                        "failed",
                        properties={
                            "error": str(e),
                            "trace": traceback.format_exc(),
                            "reason": "Invalid request error / context length",
                            **metadata,
                            "duration": round(time() - on_ticket_start_time),
                        },
                    )
                    delete_branch = True
                    raise e
                except AssistantRaisedException as e:
                    if ticket_progress is not None:
                        ticket_progress.status = TicketProgressStatus.ERROR
                        ticket_progress.error_message = f"Sweep raised an error with the following message: {e.message}. Feel free to add more details to the issue descript for Sweep to better address it, or alternatively, reach out to Kevin or William for help at https://discord.gg/sweep."
                        ticket_progress.save()

                    logger.exception(e)
                    edit_sweep_comment(
                        f"Sweep raised an error with the following message:\n{blockquote(e.message)}",
                        -1,
                    )
                    log_error(
                        is_paying_user,
                        is_consumer_tier,
                        username,
                        issue_url,
                        "Workflow",
                        str(e) + "\n" + traceback.format_exc(),
                        priority=1,
                    )
                    raise e
                except Exception as e:
                    ticket_progress.status = TicketProgressStatus.ERROR
                    ticket_progress.error_message = f"Internal server error: {str(e)}. Feel free to add more details to the issue descript for Sweep to better address it, or alternatively, reach out to Kevin or William for help at https://discord.gg/sweep."
                    ticket_progress.save()

                    logger.error(traceback.format_exc())
                    logger.error(e)
                    # title and summary are defined elsewhere
                    if len(title + summary) < 60:
                        edit_sweep_comment(
                            (
                                "I'm sorry, but it looks like an error has occurred due to"
                                + " a planning failure. Feel free to add more details to the issue description"
                                + " so Sweep can better address it. Alternatively, reach out to Kevin or William for help at"
                                + " https://discord.gg/sweep."
                            ),
                            -1,
                        )
                    else:
                        edit_sweep_comment(
                            (
                                "I'm sorry, but it looks like an error has occurred due to"
                                + " a planning failure. Feel free to add more details to the issue description"
                                + " so Sweep can better address it. Alternatively, reach out to Kevin or William for help at"
                                + " https://discord.gg/sweep."
                            ),
                            -1,
                        )
                    log_error(
                        is_paying_user,
                        is_consumer_tier,
                        username,
                        issue_url,
                        "Workflow",
                        str(e) + "\n" + traceback.format_exc(),
                        priority=1,
                    )
                    raise e
                else:
                    try:
                        fire_and_forget_wrapper(remove_emoji)(content_to_delete="eyes")
                        fire_and_forget_wrapper(add_emoji)("rocket")
                    except SystemExit:
                        raise SystemExit
                    except Exception as e:
                        logger.error(e)

                if delete_branch:
                    try:
                        if pull_request.branch_name.startswith("sweep"):
                            repo.get_git_ref(
                                f"heads/{pull_request.branch_name}"
                            ).delete()
                        else:
                            raise Exception(
                                f"Branch name {pull_request.branch_name} does not start with sweep/"
                            )
                    except SystemExit:
                        raise SystemExit
                    except Exception as e:
                        logger.error(e)
                        logger.error(traceback.format_exc())
                        logger.info("Deleted branch", pull_request.branch_name)
            except Exception as e:
                posthog.capture(
                    username,
                    "failed",
                    properties={
                        **metadata,
                        "error": str(e),
                        "trace": traceback.format_exc(),
                        "duration": round(time() - on_ticket_start_time),
                    },
                )
                raise e
        posthog.capture(
            username,
            "success",
            properties={**metadata, "duration": round(time() - on_ticket_start_time)},
        )
        logger.info("on_ticket success in " + str(round(time() - on_ticket_start_time)))
        return {"success": True}


def handle_sandbox_mode(
    title, repo_full_name, repo, ticket_progress, edit_sweep_comment
):
    logger.info("Running in sandbox mode")
    sweep_bot = SweepBot(repo=repo, ticket_progress=ticket_progress)
    logger.info("Getting file contents")
    file_name = title.split(":")[1].strip()
    file_contents = sweep_bot.get_contents(file_name).decoded_content.decode("utf-8")
    try:
        ext = file_name.split(".")[-1]
    except Exception:
        ext = ""
    file_contents.replace("```", "\`\`\`")
    sha = repo.get_branch(repo.default_branch).commit.sha
    permalink = f"https://github.com/{repo_full_name}/blob/{sha}/{file_name}#L1-L{len(file_contents.splitlines())}"
    logger.info("Running sandbox")
    edit_sweep_comment(
        f"Running sandbox for {file_name}. Current Code:\n\n{permalink}",
        1,
    )
    updated_contents, sandbox_response = sweep_bot.check_sandbox(
        file_name, file_contents, []
    )
    logger.info("Sandbox finished")
    logs = (
        (
            "<br/>"
            + create_collapsible(
                "Sandbox logs",
                blockquote(
                    "\n\n".join(
                        [
                            create_collapsible(
                                f"<code>{output}</code> {i + 1}/{len(sandbox_response.outputs)} {format_sandbox_success(sandbox_response.success)}",
                                f"<pre>{clean_logs(output)}</pre>",
                                i == len(sandbox_response.outputs) - 1,
                            )
                            for i, output in enumerate(sandbox_response.outputs)
                            if len(sandbox_response.outputs) > 0
                        ]
                    )
                ),
                opened=True,
            )
        )
        if sandbox_response
        else ""
    )

    updated_contents = updated_contents.replace("```", "\`\`\`")
    diff = generate_diff(file_contents, updated_contents).replace("```", "\`\`\`")
    diff_display = (
        f"Updated Code:\n\n```{ext}\n{updated_contents}```\nDiff:\n```diff\n{diff}\n```"
        if diff
        else f"Sandbox made no changes to {file_name} (formatters were not configured or Sweep didn't make changes)."
    )

    edit_sweep_comment(
        f"{logs}\n{diff_display}",
        2,
    )
    edit_sweep_comment("N/A", 3)
    logger.info("Sandbox comments updated")


def get_branch_diff_text(repo, branch, base_branch=None):
    base_branch = base_branch or SweepConfig.get_branch(repo)
    comparison = repo.compare(base_branch, branch)
    file_diffs = comparison.files

    pr_diffs = []
    for file in file_diffs:
        diff = file.patch
        if (
            file.status == "added"
            or file.status == "modified"
            or file.status == "removed"
        ):
            pr_diffs.append((file.filename, diff))
        else:
            logger.info(
                f"File status {file.status} not recognized"
            )  # TODO(sweep): We don't handle renamed files
    return "\n".join([f"{filename}\n{diff}" for filename, diff in pr_diffs])


def get_payment_messages(chat_logger: ChatLogger):
    if chat_logger:
        is_paying_user = chat_logger.is_paying_user()
        is_consumer_tier = chat_logger.is_consumer_tier()
        use_faster_model = chat_logger.use_faster_model()
    else:
        is_paying_user = True
        is_consumer_tier = False
        use_faster_model = False

    tracking_id = chat_logger.data["tracking_id"] if MONGODB_URI is not None else None

    # Find the first comment made by the bot
    tickets_allocated = 5
    if is_consumer_tier:
        tickets_allocated = 15
    if is_paying_user:
        tickets_allocated = 500
    purchased_ticket_count = (
        chat_logger.get_ticket_count(purchased=True) if chat_logger else 0
    )
    ticket_count = (
        max(tickets_allocated - chat_logger.get_ticket_count(), 0)
        + purchased_ticket_count
        if chat_logger
        else 999
    )
    daily_ticket_count = (
        (3 - chat_logger.get_ticket_count(use_date=True) if not use_faster_model else 0)
        if chat_logger
        else 999
    )

    model_name = "GPT-4"
    single_payment_link = "https://buy.stripe.com/00g3fh7qF85q0AE14d"
    pro_payment_link = "https://buy.stripe.com/00g5npeT71H2gzCfZ8"
    daily_message = (
        f" and {daily_ticket_count} for the day"
        if not is_paying_user and not is_consumer_tier
        else ""
    )
    user_type = "💎 <b>Sweep Pro</b>" if is_paying_user else "⚡ <b>Sweep Basic Tier</b>"
    gpt_tickets_left_message = (
        f"{ticket_count} GPT-4 tickets left for the month"
        if not is_paying_user
        else "unlimited GPT-4 tickets"
    )
    purchase_message = f"<br/><br/> For more GPT-4 tickets, visit <a href={single_payment_link}>our payment portal</a>. For a one week free trial, try <a href={pro_payment_link}>Sweep Pro</a> (unlimited GPT-4 tickets)."
    payment_message = (
        f"{user_type}: I used {model_name} to create this ticket. You have {gpt_tickets_left_message}{daily_message}. (tracking ID: <code>{tracking_id}</code>)"
        + (purchase_message if not is_paying_user else "")
    )
    payment_message_start = (
        f"{user_type}: I'm using {model_name}. You have {gpt_tickets_left_message}{daily_message}. (tracking ID: <code>{tracking_id}</code>)"
        + (purchase_message if not is_paying_user else "")
    )

    return payment_message, payment_message_start
